# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pathlib import Path

try:
    import yaml
    HAVE_YAML = True
except ImportError:
    HAVE_YAML = False

try:
    from junit_xml import TestSuite, TestCase, to_xml_report_string
    HAVE_JUNIT_XML = True
except ImportError:
    HAVE_JUNIT_XML = False

from loki.ir import Node
from loki.lint.utils import get_filename_from_parent, is_rule_disabled, get_location_hash
from loki.logging import logger, error
from loki.module import Module
from loki.sourcefile import Sourcefile
from loki.subroutine import Subroutine
from loki.tools import filehash

__all__ = [
    'ProblemReport', 'RuleReport', 'FileReport', 'Reporter',
    'GenericHandler', 'DefaultHandler', 'ViolationFileHandler',
    'JunitXmlHandler', 'LazyTextfile'
]


class ProblemReport:
    """
    Data type to represent a problem reported for a node in the IR

    Parameters
    ----------
    msg : str
        The message describing the problem.
    location : :any:`Sourcefile` or :any:`Module` or :any:`Subroutine` or :any:`Node`
        The IR component in which the problem exists.
    """

    def __init__(self, msg, location):
        self.msg = msg
        self.location = location


class RuleReport:
    """
    Container type to collect all individual problems reported by a rule

    All :class:`RuleReport` instances that belong to a file are
    collected in a :class:`FileReport`.

    Parameters
    ----------
    rule : :any:`GenericRule`
        The rule that generated the report
    reports : list of :any:`ProblemReport`, optional
        List of problem reports for this rule
    disabled : bool or list, optional
        Flag to disable reporting of this rule. If set to `True`, no violations will
        be reported. If a list of hashes is provided, the first line of a violation
        will be compared against these hashes and not reported if found.
    """

    def __init__(self, rule, reports=None, disabled=None):
        self.rule = rule
        self.problem_reports = reports or []
        self.disabled = disabled or []
        self.elapsed_sec = 0.

    def add(self, msg, location):
        """
        Convenience function to append a problem report to the list of problems
        reported by the rule.

        Parameters
        ----------
        msg : str
            The message describing the problem.
        location : :any:`Sourcefile` or :any:`Module` or :any:`Subroutine` or :any:`Node`
            The IR node or expression node in which the problem exists.
        """
        if self.disabled is True:
            return
        if not isinstance(location, (Sourcefile, Module, Subroutine, Node)):
            raise TypeError(f'Invalid type for report location: {type(location).__name__}')
        if not is_rule_disabled(location, self.rule.identifiers(), self.disabled):
            self.problem_reports.append(ProblemReport(msg, location))


class FileReport:
    """
    Container type to collect all rule reports for a file

    Parameters
    ----------
    filename : str
        The filename of the file the report is for
    hash : str, optional
        Provide a hash for the file's content to identify the file version
    reports : list, optional
        List of :py:class:`RuleReport`.
    """

    def __init__(self, filename, hash=None, reports=None):  # pylint: disable=redefined-builtin
        self.filename = filename
        self.hash = hash or filehash(Path(filename).read_text())
        self.reports = reports or []

    def add(self, rule_report):
        """
        Append a rule report to the list of reports.

        Parameters
        -----------
        rule_report : :any:`RuleReport`
            The report to be stored.
        """
        if not isinstance(rule_report, RuleReport):
            raise TypeError(f'{type(rule_report)} given, {RuleReport} expected')
        self.reports.append(rule_report)

    @property
    def fixable_reports(self):
        """
        Yield only those rule reports that belong to a rule that can be fixed.
        """
        fixable_reports = [report for report in self.reports
                           if report.rule.fixable and report.problem_reports]
        return fixable_reports


class Reporter:
    """
    Manager for problem reports and their handler.

    It collects file reports and feeds them to all available handlers to generate
    their individual reporting pieces.
    Note that this processing of reports happens immediately when adding a new file
    report for two reasons:

    #. Enable immediate output functionality (i.e., being able to print problems
       as soon as they are detected and not only at the very end of a (lengthy)
       multi file parser run.
    #. To allow parallel processing. The location of problem reports is not
       pickable and thus they need to be processed into a pickable form.

    The class maintains a `dict` in which a list of reports is stored for each handler.
    In a parallel setting, this needs to be initialized explicitly to enable thread
    safe data structures by calling `init_parallel()`.

    Parameters
    ----------
    list handlers : list of :any:`GenericHandler`, optional
        The enabled handlers. If none given, :any:`DefaultHandler` will be used.
    """

    def __init__(self, handlers=None):
        if not handlers:
            handlers = [DefaultHandler()]
        self.handlers_reports = {handler: [] for handler in handlers}

    def init_parallel(self, manager):
        """
        Additional initialization step when using the reporter in a parallel setting.

        Parameters
        ----------
        manager : :any:`multiprocessing.Manager`
            The multiprocessing manager that should be used to create thread safe data structures.
        """
        parallel_reports = manager.dict()
        for handler, reports in self.handlers_reports.items():
            parallel_reports[handler] = manager.list(reports)
        self.handlers_reports = parallel_reports

    def add_file_report(self, file_report):
        """
        Process a file report in all handlers and store the results.

        :param :py:class:`FileReport` file_report: the file report to be processed.
        """
        if not isinstance(file_report, FileReport):
            raise TypeError(f'{type(file_report)} given, {FileReport} expected')
        for handler, reports in self.handlers_reports.items():
            reports.append(handler.handle(file_report))

    def add_file_error(self, filename, rule, msg):
        """
        Create a file report with a single problem reported and add it.

        This is a convenience function that can be used, e.g., to report a failing rule
        or other problems with a certain file.

        Parameters
        ----------
        filename : str
            The file name of the corresponding file.
        rule : :any:`GenericRule`
            The rule that exposed the problem or `None`.
        msg : str
            A description of the problem.
        """
        problem_report = ProblemReport(msg, None)
        rule_report = RuleReport(rule, reports=[problem_report])
        file_report = FileReport(filename, reports=[rule_report])
        self.add_file_report(file_report)

    def output(self):
        """
        Call the `output` function for all reports on every handler.
        """
        for handler, reports in self.handlers_reports.items():
            handler.output(reports)


class GenericHandler:
    """
    Base class for report handler.

    Parameters
    ----------
    basedir : str, optional
        Base directory path relative to which file paths are given.
    """

    def __init__(self, basedir=None):
        self.basedir = basedir

    def get_relative_filename(self, filename):
        if filename and self.basedir:
            try:
                filename = Path(filename).relative_to(self.basedir)
            except ValueError:
                pass
        return filename


    def format_location(self, filename, location):
        """
        Create a string representation of the location given in a `ProblemReport`.

        For a given location it tries to determine:
            - the file name (if not given)
            - the source line
            - the name of the scope (i.e., enclosing subroutine or module)

        Parameters
        ----------
        filename : str
            The file name of the source file.
        location : :any:`Node` or :any:`Subroutine` or :any:`Sourcefile` or :any:`Module`
            The AST node that triggered the problem report.

        Returns
        -------
        str
            The formatted string in the form
            "<filename> (l. <line(s)>) [in routine/module ...]"
        """
        if not filename:
            filename = get_filename_from_parent(location) or ''
        filename = self.get_relative_filename(filename)

        source = getattr(location, '_source', getattr(location, 'source', None))
        if source is not None:
            line = f' (l. {source.lines[0]})'
        else:
            line = ''

        if isinstance(location, Subroutine):
            scope = f' in routine "{location.name}"'
        elif isinstance(location, Module):
            scope = f' in module "{location.name}"'
        else:
            scope = ''
        return f'{filename}{line}{scope}'

    def handle(self, file_report):  # pylint: disable=unused-argument
        """
        Handle the given :attr:`file_report`.

        This routine has to be implemented by the handler class.
        It should either print/save the report immediately or return a picklable
        object that is later to be printed/saved via :meth:`output`.

        Note that the only requirement is that :meth:`handle` and
        :meth:`output` are compatible in the sense that a list of objects
        returned by :meth:`handle` can be processed by :meth:`output`.
        """
        raise NotImplementedError()

    def output(self, handler_reports):
        """
        Output the list of report objects created by :meth:`handle`.
        """
        raise NotImplementedError()


class DefaultHandler(GenericHandler):
    """
    The default report handler for command line output of problems.

    Parameters
    ----------
    target : optional
        The output destination as a callback. Will be called with a string.
        Defaults to :attr:`loki.logging.logger.warning`
    immediate_output : bool, optional
        Print problems immediately if `True`, otherwise
        collect messages and print when calling `output()`. Defaults to `True`
    basedir : str, optional
        Base directory path relative to which file paths are given.
    """

    fmt_string = '{rule}: {location} - {msg}'

    def __init__(self, target=logger.warning, immediate_output=True, basedir=None):
        super().__init__(basedir)
        self.target = target
        self.immediate_output = immediate_output

    def handle(self, file_report):
        """
        Creates a string output of all problem reports and (by default) prints them
        immediately to `target`.

        Parameters
        ----------
        file_report : :any:`FileReport`
            The file report to be processed.

        Returns
        -------
        list of str
            The list of problem report strings.
        """
        filename = file_report.filename
        reports_list = []
        for rule_report in file_report.reports:
            rule = rule_report.rule.__name__
            if hasattr(rule_report.rule, 'docs') and rule_report.rule.docs:
                if 'id' in rule_report.rule.docs:
                    rule = f'[{rule_report.rule.docs["id"]}] {rule}'
            for problem in rule_report.problem_reports:
                location = self.format_location(filename, problem.location)
                msg = self.fmt_string.format(rule=rule, location=location, msg=problem.msg)
                if self.immediate_output:
                    self.target(msg)
                reports_list.append(msg)
        return reports_list

    def output(self, handler_reports):
        """
        Print all reports to `target` if `immediate_output` is disabled.

        Parameters
        ----------
        handler_reports : list of list of str
            The list of lists of reports.
        """
        if not self.immediate_output:
            for reports in handler_reports:
                for report in reports:
                    self.target(report)


class ViolationFileHandler(GenericHandler):
    """
    Report handler class that writes a YAML file with rules violated per file

    The content of the YAML file has the form

    .. code-block:: yaml

        'path/to/my/file.F90':
            filehash: 'abc123'
            rules:
            - SomeRule
            - SomeOtherRule

    This YAML file can be included into a linter config file to disable reporting
    of these rules on that version of the file (source modifications identified by
    :any:`filehash`) in future runs.

    An alternative format of this file is the following

    .. code-block:: yaml

        'path/to/my/file.F90':
            rules:
            - Some Rule:
              - <line hash>
              - <line hash>
            - SomeOtherRule:
              - <line hash>

    This will disable reporting violations of given rules for the specified file,
    as long as the line hash matches one of the listed line hashes. This file format
    can be generated by specifying :data:`use_line_hashes`.

    Parameters
    ----------
    target :
        The output destination
    basedir : str, optional
        Base directory path relative to which file paths are given.
    use_line_hashes : bool, optional
        Disable rule violations per line
    """
    def __init__(self, target=logger.warning, basedir=None, use_line_hashes=False):
        if not HAVE_YAML:
            error('Pyyaml is not available')
            raise RuntimeError

        super().__init__(basedir)
        self.target = target
        self.use_line_hashes = use_line_hashes

    def handle(self, file_report):
        """
        Create YAML block for this file

        Parameters
        ----------
        file_report : :any:`FileReport`
            The file report to be processed

        Returns
        -------
        str
            YAML block for this file
        """
        if self.use_line_hashes:
            violated_rules = [
                {
                    rule_report.rule.__name__: [
                        line_hash for problem_report in rule_report.problem_reports
                        if (line_hash := get_location_hash(problem_report.location))
                    ]
                }
                for rule_report in file_report.reports
                if rule_report.problem_reports
            ]
        else:
            violated_rules = [
                rule_report.rule.__name__ for rule_report in file_report.reports
                if rule_report.problem_reports
            ]

        if violated_rules:
            violations_report = {}
            violations_report['rules'] = violated_rules
            if not self.use_line_hashes:
                violations_report['filehash'] = file_report.hash
            return yaml.dump({
                str(self.get_relative_filename(file_report.filename)): violations_report
            })
        return ''

    def output(self, handler_reports):
        """
        Generate the YAML output from the list of reports.
        """
        self.target('\n'.join(report for report in handler_reports if report))


class JunitXmlHandler(GenericHandler):
    """
    Report handler class that generates JUnit-compatible XML output that can be understood
    by CI platforms such as Jenkins or Bamboo

    Parameters
    ----------
    target :
        The output destination
    basedir : str, optional
        Base directory path relative to which file paths are given.
    """

    fmt_string = '{location} - {msg}'

    def __init__(self, target=logger.warning, basedir=None):
        if not HAVE_JUNIT_XML:
            error('junit_xml is not available')
            raise RuntimeError

        super().__init__(basedir)
        self.target = target

    def handle(self, file_report):
        """
        Creates tuples of string arguments for `junit_xml.TestCase`

        Parameters
        ----------
        file_report : :any:`FileReport`
            The file report to be processed

        Returns
        -------
        tuple(str, list)
            Tuples of the form ``(filename, [(kwargs, messages)])`` with :attr:`kwargs`
            being the constructor arguments for `junit_xml.TestCase` and
            :attr:`messages` a list of strings.
        """
        filename = file_report.filename
        classname = str(Path(filename).with_suffix(''))
        test_cases = []
        for rule_report in file_report.reports:
            kwargs = {'name': rule_report.rule.__name__, 'classname': classname,
                      'allow_multiple_subelements': True, 'elapsed_sec': rule_report.elapsed_sec}
            messages = []
            for problem in rule_report.problem_reports:
                location = self.format_location(filename, problem.location)
                msg = self.fmt_string.format(location=location, msg=problem.msg)
                messages.append(msg)
            test_cases.append((kwargs, messages))
        return (filename, test_cases)

    def output(self, handler_reports):
        """
        Generate the XML output from the list of reports.
        """
        testsuites = []
        for filename, tc_args in handler_reports:
            testcases = []
            for kwargs, messages in tc_args:
                testcase = TestCase(**kwargs)
                for msg in messages:
                    testcase.add_failure_info(msg)
                testcases.append(testcase)
            testsuites.append(TestSuite(filename, testcases))
        xml_string = to_xml_report_string(testsuites)
        self.target(xml_string)


class LazyTextfile:
    """
    Helper class to encapsulate opening and writing to a file.

    This exists because opening the file immediately and then passing
    its ``write`` function to a :any:`GenericHandler` makes it
    impossible to pickle it afterwards, which would make parallel
    execution infeasible.

    Instead of creating a more complicated interface for the handlers
    we opted for this way of a just-in-time/lazy file handler.

    The file is opened automatically for writing when calling :meth:`write`
    for the first time, and closed when the object is garbage collected.

    Parameters
    ----------
    filename : str
        The filename of the output file
    """

    def __init__(self, filename):
        self.file_name = Path(filename)
        self.file_handle = None

    def _check_open(self):
        """
        Check if the file is open already, otherwise open it
        """
        if not self.file_handle:
            self.file_handle = self.file_name.open(mode='w')  # pylint: disable=consider-using-with

    def __del__(self):
        if self.file_handle:
            self.file_handle.close()
            self.file_handle = None

    def write(self, msg):
        """
        Write the given :data:`msg` to the file

        The file is opened first, unless it is open already
        """
        self._check_open()
        self.file_handle.write(msg)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Various tests for interprocedural analysis features in Loki
"""

import pytest

from loki import (
    Sourcefile, FindNodes, FindInlineCalls, CallStatement, IntLiteral
)
from loki.frontend import available_frontends


@pytest.mark.parametrize('frontend', available_frontends())
def test_ipa_call_statement_arg_iter(frontend, tmp_path):
    """
    Test that :any:`CallStatement.arg_iter` works as expected
    """
    fcode_caller = """
subroutine caller
    use some_mod, only: callee
    implicit none
    integer :: arg1, arg3(10)
    real :: arg2
    call callee(arg1, arg2, arg3, 4)
end subroutine caller
    """.strip()

    fcode_callee = """
module some_mod
    implicit none
contains
    subroutine callee(var, VAR2, arr, val)
        integer, intent(inout) :: var
        real, intent(in) :: var2
        integer, intent(in) :: arr(:)
        integer, intent(in) :: val
    end subroutine callee
end module some_mod
    """.strip()

    callee_source = Sourcefile.from_source(fcode_callee, frontend=frontend, xmods=[tmp_path])
    caller_source = Sourcefile.from_source(
        fcode_caller, frontend=frontend, xmods=[tmp_path],
        definitions=callee_source.definitions
    )

    callee = callee_source['callee']
    caller = caller_source['caller']

    calls = FindNodes(CallStatement).visit(caller.body)
    assert len(calls) == 1
    arg_iter = list(calls[0].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('arr(:)', 'arg3'), ('val', '4')
    ]

    for kernel_arg, caller_arg in calls[0].arg_iter():
        assert kernel_arg.scope is callee
        assert isinstance(caller_arg, IntLiteral) or caller_arg.scope is caller


@pytest.mark.parametrize('frontend', available_frontends())
def test_ipa_call_statement_arg_iter_optional(frontend, tmp_path):
    """
    Test that :any:`CallStatement.arg_iter` works as expected with optional arguments
    """
    fcode_caller = """
subroutine caller
    use some_mod, only: callee
    implicit none
    integer :: arg1, arg3(10)
    real :: arg2
    call callee(arg1, arg2, arg3, 4)
    call callee(arg1, arg2, VAL=4, arr=arg3, opt2=1)
    call callee(arg1, arg2, arg3, 4, 1, 2)
end subroutine caller
    """.strip()

    fcode_callee = """
module some_mod
    implicit none
contains
    subroutine callee(var, VAR2, arr, val, OPT1, opt2)
        integer, intent(inout) :: var
        real, intent(in) :: var2
        integer, intent(in) :: arr(:)
        integer, intent(in) :: val
        integer, intent(in), optional :: OPT1, opt2
    end subroutine callee
end module some_mod
    """.strip()

    callee_source = Sourcefile.from_source(fcode_callee, frontend=frontend, xmods=[tmp_path])
    caller_source = Sourcefile.from_source(
        fcode_caller, frontend=frontend, xmods=[tmp_path],
        definitions=callee_source.definitions
    )

    callee = callee_source['callee']
    caller = caller_source['caller']

    calls = FindNodes(CallStatement).visit(caller.body)
    assert len(calls) == 3
    arg_iter = list(calls[0].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('arr(:)', 'arg3'), ('val', '4')
    ]
    arg_iter = list(calls[1].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('val', '4'), ('arr(:)', 'arg3'), ('opt2', '1')
    ]
    arg_iter = list(calls[2].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('arr(:)', 'arg3'), ('val', '4'), ('opt1', '1'), ('opt2', '2')
    ]

    for call in calls:
        for kernel_arg, caller_arg in call.arg_iter():
            assert kernel_arg.scope is callee
            assert isinstance(caller_arg, IntLiteral) or caller_arg.scope is caller


@pytest.mark.parametrize('frontend', available_frontends())
def test_ipa_inline_call_arg_iter(frontend, tmp_path):
    """
    Test that :any:`InlineCall.arg_iter` works as expected
    """
    fcode_caller = """
subroutine caller
    use some_mod, only: callee
    implicit none
    integer :: arg1, arg3(10)
    real :: arg2, ret
    ret = callee(arg1, arg2, arg3, 4)
end subroutine caller
    """.strip()

    fcode_callee = """
module some_mod
    implicit none
contains
    function callee(var, VAR2, arr, val)
        integer, intent(inout) :: var
        real, intent(in) :: var2
        integer, intent(in) :: arr(:)
        integer, intent(in) :: val
        real :: callee
    end function callee
end module some_mod
    """.strip()

    callee_source = Sourcefile.from_source(fcode_callee, frontend=frontend, xmods=[tmp_path])
    caller_source = Sourcefile.from_source(
        fcode_caller, frontend=frontend, xmods=[tmp_path],
        definitions=callee_source.definitions
    )

    callee = callee_source['callee']
    caller = caller_source['caller']

    calls = list(FindInlineCalls().visit(caller.body))
    assert len(calls) == 1
    arg_iter = list(calls[0].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('arr(:)', 'arg3'), ('val', '4')
    ]

    for kernel_arg, caller_arg in calls[0].arg_iter():
        assert kernel_arg.scope is callee
        assert isinstance(caller_arg, IntLiteral) or caller_arg.scope is caller


@pytest.mark.parametrize('frontend', available_frontends())
def test_ipa_inline_call_arg_iter_optional(frontend, tmp_path):
    """
    Test that :any:`InlineCall.arg_iter` works as expected
    """
    fcode_caller = """
subroutine caller
    use some_mod, only: callee
    implicit none
    integer :: arg1, arg3(10)
    real :: arg2, ret
    ret = callee(arg1, arg2, arg3, 4)
    ret = ret + callee(arg1, ARR=arg3, var2=arg2, opt2=2, val=4)
    ret = ret + callee(arg1, arg2, arg3, 4, 1, 2)
end subroutine caller
    """.strip()

    fcode_callee = """
module some_mod
    implicit none
contains
    function callee(var, VAR2, arr, val, opt1, OPT2)
        integer, intent(inout) :: var
        real, intent(in) :: var2
        integer, intent(in) :: arr(:)
        integer, intent(in) :: val
        integer, intent(in) :: opt1, OPT2
        real :: callee
    end function callee
end module some_mod
    """.strip()

    callee_source = Sourcefile.from_source(fcode_callee, frontend=frontend, xmods=[tmp_path])
    caller_source = Sourcefile.from_source(
        fcode_caller, frontend=frontend, xmods=[tmp_path],
        definitions=callee_source.definitions
    )

    callee = callee_source['callee']
    caller = caller_source['caller']

    calls = list(FindInlineCalls(unique=False).visit(caller.body))
    assert len(calls) == 3
    arg_iter = list(calls[0].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('arr(:)', 'arg3'), ('val', '4')
    ]
    arg_iter = list(calls[1].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('arr(:)', 'arg3'), ('var2', 'arg2'), ('opt2', '2'), ('val', '4')
    ]
    arg_iter = list(calls[2].arg_iter())
    assert arg_iter == [
        ('var', 'arg1'), ('var2', 'arg2'), ('arr(:)', 'arg3'), ('val', '4'), ('opt1', '1'), ('opt2', '2')
    ]

    for call in calls:
        for kernel_arg, caller_arg in call.arg_iter():
            assert kernel_arg.scope is callee
            assert isinstance(caller_arg, IntLiteral) or caller_arg.scope is caller



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from loki.ir import (
    nodes as ir, Transformer, FindVariables, SubstituteExpressions
)
from loki.subroutine import Subroutine
from loki.expression import (
    symbols as sym, parse_expr, ceil_division, iteration_index
)

__all__ = ['split_loop', 'block_loop_arrays']


class LoopSplittingVariables:
    """
    This class holds the loop splitting variables, e.g. outer loop block sizes and iteration
    bounds. It also holds the original loop variable of the inner loop.
    """

    def __init__(self, loop_var: sym.Variable, block_size):
        self._loop_var = loop_var
        # self._splitting_vars = splitting_vars
        self._splitting_vars = (loop_var.clone(name=loop_var.name + "_loop_block_size",
                                               type=loop_var.type.clone(parameter=True,
                                                                        initial=sym.IntLiteral(
                                                                            block_size))),
                                loop_var.clone(name=loop_var.name + "_loop_num_blocks"),
                                loop_var.clone(name=loop_var.name + "_loop_block_idx"),
                                loop_var.clone(name=loop_var.name + "_loop_local"),
                                loop_var.clone(name=loop_var.name + "_loop_iter_num"),
                                loop_var.clone(name=loop_var.name + "_loop_block_start"),
                                loop_var.clone(name=loop_var.name + "_loop_block_end")
                                )

    @property
    def loop_var(self):
        return self._loop_var

    @property
    def block_size(self):
        return self._splitting_vars[0]

    @property
    def num_blocks(self):
        return self._splitting_vars[1]

    @property
    def block_idx(self):
        return self._splitting_vars[2]

    @property
    def inner_loop_var(self):
        return self._splitting_vars[3]

    @property
    def iter_num(self):
        return self._splitting_vars[4]

    @property
    def block_start(self):
        return self._splitting_vars[5]

    @property
    def block_end(self):
        return self._splitting_vars[6]

    @property
    def splitting_vars(self):
        return self._splitting_vars


def split_loop(routine: Subroutine, loop: ir.Loop, block_size: int):
    """
    Blocks a loop by splitting it into an outer loop and inner loop of size `block_size`.

    Parameters
    ----------
    routine: :any:`Subroutine`
        Subroutine object containing the loop. New variables introduced in the
        loop splitting will be declared in the body of routine.
    loop: :any:`Loop`
        Loop to be split.
    block_size: int
        inner loop size (size of blocking blocks)
    """

    # loop splitting variable declarations
    splitting_vars = LoopSplittingVariables(loop.variable, block_size)
    routine.variables += splitting_vars.splitting_vars

    # block index calculations
    blocking_body = (
        ir.Assignment(splitting_vars.block_start,
                      parse_expr(
                          f"({splitting_vars.block_idx} - 1) * {splitting_vars.block_size} + 1",
                          scope=routine)
                      ),
        ir.Assignment(splitting_vars.block_end,
                      sym.InlineCall(sym.DeferredTypeSymbol('MIN', scope=routine),
                                     parameters=(sym.Product(children=(
                                         splitting_vars.block_idx, splitting_vars.block_size)),
                                                 loop.bounds.upper))
                      ))

    # Outer loop blocking variable assignments
    loop_range = loop.bounds
    block_loop_inits = (
        ir.Assignment(splitting_vars.num_blocks,
                      ceil_division(loop_range.num_iterations,
                                    splitting_vars.block_size)),
    )

    # Inner loop
    iteration_nums = (
        ir.Assignment(splitting_vars.iter_num,
                      parse_expr(
                          f"{splitting_vars.block_start}+{splitting_vars.inner_loop_var}-1"),
                      scope=routine),
        ir.Assignment(loop.variable,
                      iteration_index(splitting_vars.iter_num, loop_range))
    )
    inner_loop = ir.Loop(variable=splitting_vars.inner_loop_var, body=iteration_nums + loop.body,
                         bounds=sym.LoopRange(
                             (sym.IntLiteral(1), parse_expr(
                                 f"{splitting_vars.block_end} - {splitting_vars.block_start} + 1",
                                 scope=routine))))

    #  Outer loop bounds + body
    outer_loop = ir.Loop(variable=splitting_vars.block_idx, body=blocking_body + (inner_loop,),
                         bounds=sym.LoopRange((sym.IntLiteral(1), splitting_vars.num_blocks)))
    change_map = {loop: block_loop_inits + (outer_loop,)}
    Transformer(change_map, inplace=True).visit(routine.body)
    return splitting_vars, inner_loop, outer_loop


def blocked_shape(a: sym.Array, blocking_indices, block_size):
    """
    calculates the dimensions for a blocked version of the array.
    """
    shape = tuple(
        sym.IntLiteral(block_size) if isinstance(dim, sym.Scalar) and any(
            bidx in dim for bidx in blocking_indices) else dim for dim
        in a.shape)
    return shape


def blocked_type(a: sym.Array):
    return a.type.clone(intent=None)


def replace_indices(dimensions, indices: list, replacement_index):
    """
    Returns a new dimension object with all occurences of indices changed to replacement_index.

    Parameters
    ----------
    dimensions:
        Symbolic representation of dimensions or indices.
    indices: list of `Variable`s
        that will be replaced in the new :any:`Dimension` object.
    replacement_index: :any:`Expression`
        replacement for the indices changed.

    Returns
    -------
    """
    dims = tuple(
        replacement_index if isinstance(dim, sym.Scalar) and any(
            blocking_var in dim for blocking_var in indices) else dim for dim
        in dimensions)
    return dims


def block_loop_arrays(routine: Subroutine, splitting_vars, inner_loop: ir.Loop,
                      outer_loop: ir.Loop, blocking_indices):
    """
    Replaces arrays inside the inner loop with blocked counterparts.

    This routine declares array variables to hold the blocks of the arrays used inside
    the loop and replaces array variables inside the loop with their blocked counterparts.
    An array is blocked with the leading dimensions

    Parameters
    ----------
    routine : Subroutine
        routine in which the blocking variables should be added.
    blocking_indices: list of  :any:`Variable`
        list of the index variables that arrays inside the loop should be blocked by.
    inner_loop: :any:`Loop`
        inner loop after loop splitting
    outer_loop : :any:`Loop`
        outer loop body after loop splitting
    blocking_indices : tuple or list of str
           Variable names of the indexes that should be blocked if in array
            expressions.

    """
    # Declare Blocked arrays
    arrays = tuple(var for var in FindVariables().visit(inner_loop.body) if
                   isinstance(var, sym.Array) and any(
                       bi in var for bi in blocking_indices))
    name_map = {a.name: a.name + '_block' for a in arrays}
    block_arrays = tuple(
        a.clone(name=name_map[a.name],
                dimensions=blocked_shape(a, blocking_indices, splitting_vars.block_size),
                type=blocked_type(a)) for a in arrays)
    routine.variables += block_arrays

    # Replace arrays in loop with blocked arrays and update idx
    block_array_expr = (
        a.clone(name=name_map[a.name],
                dimensions=replace_indices(a.dimensions, blocking_indices, inner_loop.variable))
        for a in arrays
    )
    SubstituteExpressions(dict(zip(arrays, block_array_expr)), inplace=True).visit(inner_loop.body)

    # memory copies
    block_range = sym.RangeIndex((splitting_vars.block_start, splitting_vars.block_end))
    local_range = sym.RangeIndex(
        (sym.IntLiteral(1),
         parse_expr(f"{splitting_vars.block_end} - {splitting_vars.block_start} + 1",
                    scope=routine)))
    # input variables
    in_vars = (a for a in arrays if a.type.intent in ('in', 'inout'))
    copyins = tuple(
        ir.Assignment(a.clone(name=name_map[a.name],
                              dimensions=replace_indices(a.dimensions, blocking_indices,
                                                         local_range)),
                      a.clone(
                          dimensions=replace_indices(a.dimensions, blocking_indices, block_range)))
        for a in in_vars)
    # output variables
    out_vars = (a for a in arrays if a.type.intent in ('out', 'inout'))
    copyouts = tuple(
        ir.Assignment(
            a.clone(dimensions=replace_indices(a.dimensions, blocking_indices, block_range)),
            a.clone(name=name_map[a.name],
                    dimensions=replace_indices(a.dimensions, blocking_indices, local_range))
        )
        for a in out_vars)
    change_map = {inner_loop: copyins + (inner_loop,) + copyouts}
    Transformer(change_map, inplace=True).visit(outer_loop)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# pylint: disable=too-many-lines

from collections import deque
from pathlib import Path
import re
import networkx as nx
import pytest

from loki import (
    Sourcefile, Subroutine, as_tuple, RawSource, TypeDef,
    Scalar, ProcedureSymbol
)
from loki.batch import (
    FileItem, ModuleItem, ProcedureItem, TypeDefItem,
    ProcedureBindingItem, ExternalItem, InterfaceItem, SGraph,
    SchedulerConfig, ItemFactory
)
from loki.frontend import HAVE_FP, REGEX, RegexParserClass
from loki.ir import nodes as ir


pytestmark = pytest.mark.skipif(not HAVE_FP, reason='Fparser not available')


@pytest.fixture(scope='module', name='here')
def fixture_here():
    return Path(__file__).parent


@pytest.fixture(scope='module', name='testdir')
def fixture_testdir(here):
    return here.parent.parent/'tests'


@pytest.fixture(name='default_config', scope='function')
def fixture_default_config():
    """
    Default SchedulerConfig configuration with basic options.
    """
    return {
        'default': {
            'mode': 'idem',
            'role': 'kernel',
            'expand': True,
            'strict': True,
            'disable': ['abort']
        },
        'routines': []
    }


@pytest.fixture(name='comp1_expected_dependencies')
def fixture_comp1_expected_dependencies():
    return {
        '#comp1': ('header_mod', 't_mod', 't_mod#t', '#comp2', 't_mod#t%proc', 't_mod#t%no%way'),
        '#comp2': ('header_mod', 't_mod#t', 'a_mod#a', 'b_mod#b', 't_mod#t%yay%proc'),
        'a_mod#a': ('header_mod',),
        'b_mod#b': (),
        't_mod': ('tt_mod#tt',),
        't_mod#t': ('tt_mod#tt', 't_mod#t1'),
        't_mod#t1': (),
        't_mod#t%proc': ('t_mod#t_proc',),
        't_mod#t_proc': ('t_mod#t', 'a_mod#a', 't_mod#t%yay%proc'),
        't_mod#t%no%way': ('t_mod#t1%way',),
        't_mod#t%yay%proc': ('tt_mod#tt%proc',),
        't_mod#t1%way': ('t_mod#my_way',),
        't_mod#my_way': ('t_mod#t1',),
        'tt_mod#tt': (),
        'tt_mod#tt%proc': ('tt_mod#proc',),
        'tt_mod#proc': ('tt_mod#tt',),
        'header_mod': (),
    }


@pytest.fixture(name='mod_proc_expected_dependencies')
def fixture_mod_proc_expected_dependencies():
    return {
        'other_mod#mod_proc': ('tt_mod#tt', 'tt_mod#tt%proc', 'b_mod#b'),
        'tt_mod#tt': (),
        'tt_mod#tt%proc': ('tt_mod#proc',),
        'tt_mod#proc': ('tt_mod#tt',),
        'b_mod#b': ()
    }


@pytest.fixture(name='expected_dependencies')
def fixture_expected_dependencies(comp1_expected_dependencies, mod_proc_expected_dependencies):
    dependencies = {}
    dependencies.update(comp1_expected_dependencies)
    dependencies.update(mod_proc_expected_dependencies)
    return dependencies


@pytest.fixture(name='no_expected_dependencies')
def fixture_no_expected_dependencies():
    return {}


@pytest.fixture(name='file_dependencies')
def fixture_file_dependencies():
    return {
        'source/comp1.F90': (
            'source/comp2.f90',
            'module/t_mod.F90',
            'headers/header_mod.F90'
        ),
        'source/comp2.f90': (
            'module/t_mod.F90',
            'headers/header_mod.F90',
            'module/a_mod.F90',
            'module/b_mod.F90'
        ),
        'module/t_mod.F90': (
            'module/tt_mod.F90',
            'module/a_mod.F90',
        ),
        'module/tt_mod.F90': (),
        'module/a_mod.F90': (
            'headers/header_mod.F90',
        ),
        'module/b_mod.F90': (),
        'headers/header_mod.F90': ()
    }


class VisGraphWrapper:
    """
    Testing utility to parse the generated callgraph visualisation.
    """

    _re_nodes = re.compile(r'\s*\"?(?P<node>[\w%#./]+)\"? \[colo', re.IGNORECASE)
    _re_edges = re.compile(r'\s*\"?(?P<parent>[\w%#./]+)\"? -> \"?(?P<child>[\w%#./]+)\"?', re.IGNORECASE)

    def __init__(self, path):
        self.text = Path(path).read_text()

    @property
    def nodes(self):
        return list(self._re_nodes.findall(self.text))

    @property
    def edges(self):
        return list(self._re_edges.findall(self.text))


def get_item(cls, path, name, parser_classes, scheduler_config=None):
    source = Sourcefile.from_file(path, frontend=REGEX, parser_classes=parser_classes)
    if scheduler_config:
        config = scheduler_config.create_item_config(name)
    else:
        config = None
    return cls(name, source=source, config=config)


def test_file_item1(testdir, default_config):
    proj = testdir/'sources/projBatch'

    # A file with simple module that contains a single subroutine
    item = get_item(FileItem, proj/'module/a_mod.F90', 'module/a_mod.F90', RegexParserClass.EmptyClass)
    assert item.name == 'module/a_mod.F90'
    assert item.local_name == item.name
    assert item.scope_name is None
    assert not item.scope
    assert item.ir is item.source
    assert str(item) == 'loki.batch.FileItem<module/a_mod.F90>'

    # A few checks on the item comparison
    assert item == 'module/a_mod.F90'
    assert item != FileItem('some_name', source=item.source)
    assert item == FileItem(item.name, source=item.source)

    # The file is not parsed at all
    assert not item.source.definitions
    assert isinstance(item.source.ir, ir.Section)
    assert len(item.source.ir.body) == 1
    assert isinstance(item.source.ir.body[0], RawSource)

    # Querying definitions triggers a round of parsing
    assert item.definitions == (item.source['a_mod'],)
    assert len(item.source.definitions) == 1

    # Without the FileItem in the item_cache, the modules will be created as ExternalItem
    assert all(
        isinstance(_item, ExternalItem) and _item.origin_cls is ModuleItem
        for _item in item.create_definition_items(
            item_factory=ItemFactory(), config=SchedulerConfig.from_dict(default_config)
        )
    )

    # Check that external item raises an exception whenever we try to access any IR nodes
    external_item = item.create_definition_items(
        item_factory=ItemFactory(), config=SchedulerConfig.from_dict(default_config)
    )[0]

    for attr in ('ir', 'scope', 'path'):
        with pytest.raises(RuntimeError):
            getattr(external_item, attr)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 1
    assert items[0] != None  # pylint: disable=singleton-comparison  # (intentionally trigger __eq__ here)
    assert items[0].name == 'a_mod'
    assert items[0].definitions == (item.source['a'],)

    # The default behavior would be to have the ProgramUnits parsed already
    item = get_item(FileItem, proj/'module/a_mod.F90', 'module/a_mod.F90', RegexParserClass.ProgramUnitClass)
    assert item.name == 'module/a_mod.F90'
    assert item.definitions == (item.source['a_mod'],)
    assert item.ir is item.source
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 1
    assert items[0].name == 'a_mod'
    assert items[0].definitions == (item.source['a'],)


def test_file_item2(testdir):
    proj = testdir/'sources/projBatch'

    # A file with a simple module that contains a single typedef
    item = get_item(FileItem, proj/'module/t_mod.F90', 'module/t_mod.F90', RegexParserClass.ProgramUnitClass)
    assert item.name == 'module/t_mod.F90'
    assert item.definitions == (item.source['t_mod'],)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 1
    assert items[0].name == 't_mod'
    assert items[0].ir is item.source['t_mod']
    # No typedefs because not selected in parser classes
    assert not items[0].ir.typedefs
    # Calling definitions automatically further completes the source
    assert items[0].definitions == (
        items[0].ir['t_proc'],
        items[0].ir['my_way'],
        items[0].ir.typedef_map['t1'],
        items[0].ir.typedef_map['t'],
    )

    # Files don't have dependencies (direct dependencies, anyway)
    assert isinstance(item.dependencies, tuple) and not item.dependencies


def test_file_item3(testdir):
    proj = testdir/'sources/projBatch'

    # The same file but with typedefs parsed from the get-go
    item = get_item(
        FileItem, proj/'module/t_mod.F90', 'module/t_mod.F90',
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass
    )
    assert item.name == 'module/t_mod.F90'
    assert item.definitions == (item.source['t_mod'],)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 1
    assert items[0].name == 't_mod'
    assert len(items[0].ir.typedefs) == 2
    assert items[0].definitions == (
        item.source['t_proc'],
        item.source['my_way'],
        item.source['t1'],
        item.source['t'],
    )

    # Filter items when calling create_definition_items()
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert not item.create_definition_items(only=ProcedureItem, item_factory=item_factory)
    items = item.create_definition_items(only=ModuleItem, item_factory=item_factory)
    assert len(items) == 1
    assert isinstance(items[0], ModuleItem)
    assert items[0].ir == item.source['t_mod']


def test_module_item1(testdir):
    proj = testdir/'sources/projBatch'

    # A file with simple module that contains a single subroutine and has no dependencies on
    # the module level
    item = get_item(ModuleItem, proj/'module/a_mod.F90', 'a_mod', RegexParserClass.ProgramUnitClass)
    assert item.name == 'a_mod'
    assert item == 'a_mod'
    assert str(item) == 'loki.batch.ModuleItem<a_mod>'
    assert item.ir is item.source['a_mod']
    assert item.definitions == (item.source['a'],)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 1
    assert isinstance(items[0], ProcedureItem)
    assert items[0].ir == item.source['a']

    assert not item.dependencies


def test_module_item2(testdir):
    proj = testdir/'sources/projBatch'

    # A different file with a simple module that contains a single subroutine but has an import
    # dependency on the module level
    item = get_item(ModuleItem, proj/'module/b_mod.F90', 'b_mod', RegexParserClass.ProgramUnitClass)
    assert item.name == 'b_mod'
    assert item.ir is item.source['b_mod']
    assert item.definitions == (item.source['b'],)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 1
    assert isinstance(items[0], ProcedureItem)
    assert items[0].ir == item.source['b']

    dependencies = item.dependencies
    assert len(dependencies) == 1
    assert isinstance(dependencies[0], ir.Import)
    assert dependencies[0].module == 'header_mod'


def test_module_item3(testdir):
    proj = testdir/'sources/projBatch'

    # Make sure the dependencies are also found correctly if done without parsing definitions first
    item = get_item(ModuleItem, proj/'module/b_mod.F90', 'b_mod', RegexParserClass.ProgramUnitClass)
    dependencies = item.dependencies
    assert len(dependencies) == 1 and dependencies[0].module == 'header_mod'


def test_module_item4(testdir):
    proj = testdir/'sources/projInlineCalls'

    # Make sure interfaces are correctly identified as definitions
    item = get_item(ModuleItem, proj/'some_module.F90', 'some_module', RegexParserClass.ProgramUnitClass)
    definitions = item.definitions
    assert len(definitions) == 6
    assert len(item.ir.interfaces) == 1
    assert item.ir.interfaces[0] in definitions

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item

    items = item.create_definition_items(item_factory=item_factory)
    assert len(items) == 8
    assert len(set(items)) == 6
    assert 'some_module#add_args' in item_factory.item_cache
    assert isinstance(item_factory.item_cache['some_module#add_args'], InterfaceItem)
    assert item_factory.item_cache['some_module#add_args'] in items


def test_procedure_item1(testdir):
    proj = testdir/'sources/projBatch'

    # A file with a single subroutine definition that calls a routine via interface block
    item = get_item(ProcedureItem, proj/'source/comp1.F90', '#comp1', RegexParserClass.ProgramUnitClass)
    assert item.name == '#comp1'
    assert item == '#comp1'
    assert str(item) == 'loki.batch.ProcedureItem<#comp1>'
    assert item.ir is item.source['comp1']
    assert isinstance(item.definitions, tuple) and not item.definitions

    assert not item.create_definition_items(item_factory=ItemFactory())

    dependencies = item.dependencies
    assert len(dependencies) == 5
    assert isinstance(dependencies[0], ir.Import)
    assert dependencies[0].module == 't_mod'
    assert isinstance(dependencies[1], ir.Import)
    assert dependencies[1].module == 'header_mod'
    assert isinstance(dependencies[2], ir.CallStatement)
    assert dependencies[2].name == 'arg%proc'
    assert isinstance(dependencies[3], ir.CallStatement)
    assert dependencies[3].name == 'comp2'
    assert isinstance(dependencies[4], ir.CallStatement)
    assert dependencies[4].name == 'arg%no%way'

    assert item.targets == ('t_mod', 't', 'nt1', 'header_mod', 'arg%proc', 'comp2', 'arg%no%way')

    # We need to have suitable dependency modules in the cache to spawn the dependency items
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache.update({
        (i := get_item(ModuleItem, proj/path, name, RegexParserClass.ProgramUnitClass)).name: i
        for path, name in [
            ('module/t_mod.F90', 't_mod'), ('source/comp2.f90', '#comp2'), ('headers/header_mod.F90', 'header_mod')
        ]
    })

    # To ensure any existing items from the item_cache are re-used, we instantiate one for
    # the procedure binding
    # pylint: disable=unsupported-binary-operation
    t_mod_t_proc = get_item(
        ProcedureBindingItem, proj/'module/t_mod.F90', 't_mod#t%proc',
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass | RegexParserClass.DeclarationClass
    )
    item_factory.item_cache[t_mod_t_proc.name] = t_mod_t_proc

    items = item.create_dependency_items(item_factory=item_factory)
    assert items == ('t_mod', 't_mod#t', 'header_mod', 't_mod#t%proc', '#comp2', 't_mod#t%no%way')
    assert item_factory.item_cache[t_mod_t_proc.name] is t_mod_t_proc
    assert items[3] is t_mod_t_proc


def test_procedure_item2(testdir):
    proj = testdir/'sources/projBatch'

    # A file with a single subroutine definition that calls two routines via module imports
    item = get_item(ProcedureItem, proj/'source/comp2.f90', '#comp2', RegexParserClass.ProgramUnitClass)
    assert item.name == '#comp2'
    assert item.ir is item.source['comp2']
    assert isinstance(item.definitions, tuple) and not item.definitions

    item_factory = ItemFactory()
    assert not item.create_definition_items(item_factory=item_factory)

    dependencies = item.dependencies
    assert len(dependencies) == 7
    assert isinstance(dependencies[0], ir.Import)
    assert dependencies[0].module == 't_mod'
    assert isinstance(dependencies[1], ir.Import)
    assert dependencies[1].module == 'header_mod'
    assert isinstance(dependencies[2], ir.Import)
    assert dependencies[2].module == 'a_mod'
    assert isinstance(dependencies[3], ir.Import)
    assert dependencies[3].module == 'b_mod'
    assert isinstance(dependencies[4], ir.CallStatement)
    assert dependencies[4].name == 'a'
    assert isinstance(dependencies[5], ir.CallStatement)
    assert dependencies[5].name == 'b'
    assert isinstance(dependencies[6], ir.CallStatement)
    assert dependencies[6].name == 'arg%yay%proc'

    assert item.targets == (
        't_mod', 't', 'header_mod', 'k',
        'a_mod', 'a', 'b_mod', 'b', 'arg%yay%proc'
    )

    # We need to have suitable dependency modules in the cache to spawn the dependency items
    item_factory.item_cache[item.name] = item
    item_factory.item_cache.update({
        (i := get_item(ModuleItem, proj/path, name, RegexParserClass.ProgramUnitClass)).name: i
        for path, name in [
            ('module/t_mod.F90', 't_mod'), ('module/a_mod.F90', 'a_mod'),
            ('module/b_mod.F90', 'b_mod'), ('headers/header_mod.F90', 'header_mod')
        ]
    })
    items = item.create_dependency_items(item_factory=item_factory)
    assert items == ('t_mod#t', 'header_mod', 'a_mod#a', 'b_mod#b', 't_mod#t%yay%proc')

    # Does it still work if we call it again?
    assert items == item.create_dependency_items(item_factory=item_factory)


def test_procedure_item3(testdir):
    proj = testdir/'sources/projBatch'

    # A file with a single subroutine declared in a module that calls a typebound procedure
    # where the type is imported via an import statement in the module scope
    item = get_item(
        ProcedureItem, proj/'module/other_mod.F90', 'other_mod#mod_proc',
        RegexParserClass.ProgramUnitClass
    )
    dependencies = item.dependencies
    assert len(dependencies) == 3
    assert dependencies[0].module == 'tt_mod'
    assert dependencies[1].name == 'arg%proc'
    assert dependencies[2].name == 'b'

    assert item.targets == ('tt_mod', 'tt', 'arg%proc', 'b')

    item_factory = ItemFactory()
    item_factory.item_cache.update({
        item.name: item,
        'tt_mod': get_item(ModuleItem, proj/'module/tt_mod.F90', 'tt_mod', RegexParserClass.ProgramUnitClass),
        'b_mod': get_item(ModuleItem, proj/'module/b_mod.F90', 'b_mod', RegexParserClass.ProgramUnitClass)
    })
    assert item.create_dependency_items(item_factory=item_factory) == ('tt_mod#tt', 'tt_mod#tt%proc', 'b_mod#b')


def test_procedure_item4(testdir):
    proj = testdir/'sources/projBatch'

    # A routine with a typebound procedure call where the typedef is in the same module
    item = get_item(
        ProcedureItem, proj/'module/t_mod.F90', 't_mod#my_way', RegexParserClass.ProgramUnitClass
    )
    dependencies = item.dependencies
    assert len(dependencies) == 2
    assert dependencies[0].name == 't1'
    assert dependencies[1].name == 'this%way'

    assert item.targets == ('t1', 'this%way')

    item_factory = ItemFactory()
    item_factory.item_cache.update({
        item.name: item,
        't_mod': ModuleItem('t_mod', source=item.source)
    })
    items = item.create_dependency_items(item_factory=item_factory)
    assert items == ('t_mod#t1', 't_mod#t1%way')


@pytest.mark.parametrize('config,expected_dependencies,expected_targets', [
    (
        {},
        ('t_mod#t', 'header_mod', 'a_mod#a', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'k', 'a_mod', 'a', 'b_mod', 'b', 'arg%yay%proc')
    ),
    (
        {'default': {'disable': ['#a']}},
        ('t_mod#t', 'header_mod', 'a_mod#a', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'k', 'a_mod', 'a', 'b_mod', 'b', 'arg%yay%proc')
    ),
    (
        {'default': {'disable': ['a']}},
        ('t_mod#t', 'header_mod', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'k', 'a_mod', 'b_mod', 'b', 'arg%yay%proc')
    ),
    (
        {'default': {'disable': ['a', 'a_mod']}},
        ('t_mod#t', 'header_mod', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'k', 'b_mod', 'b', 'arg%yay%proc'),
    ),
    (
        {'default': {'disable': ['a_mod#a']}},
        ('t_mod#t', 'header_mod', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'k', 'a_mod', 'b_mod', 'b', 'arg%yay%proc')
    ),
    (
        {'default': {'disable': ['a_mod']}},
        ('t_mod#t', 'header_mod', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'k', 'b_mod', 'b', 'arg%yay%proc')
    ),
    (
        {'default': {'disable': ['t%yay%proc']}},
        ('t_mod#t', 'header_mod', 'a_mod#a', 'b_mod#b'),
        ('t_mod', 't', 'header_mod', 'k', 'a_mod', 'a', 'b_mod', 'b')
    ),
    (
        {'default': {'disable': ['t_mod#t%yay%proc']}},
        ('t_mod#t', 'header_mod', 'a_mod#a', 'b_mod#b'),
        ('t_mod', 't', 'header_mod', 'k', 'a_mod', 'a', 'b_mod', 'b')
    ),
    (
        {'default': {'disable': ['t_mod#t']}},
        ('header_mod', 'a_mod#a', 'b_mod#b'),
        ('t_mod', 'header_mod', 'k', 'a_mod', 'a', 'b_mod', 'b')
    ),
    (
        {'default': {'disable': ['t_mod']}},
        ('header_mod', 'a_mod#a', 'b_mod#b'),
        ('header_mod', 'k', 'a_mod', 'a', 'b_mod', 'b')
    ),
    (
        {'default': {'disable': ['header_mod']}},
        ('t_mod#t', 'a_mod#a', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'a_mod', 'a', 'b_mod', 'b', 'arg%yay%proc')
    ),
    (
        {'default': {'disable': ['k']}},
        ('t_mod#t', 'a_mod#a', 'b_mod#b', 't_mod#t%yay%proc'),
        ('t_mod', 't', 'header_mod', 'a_mod', 'a', 'b_mod', 'b', 'arg%yay%proc')
    ),
])
def test_procedure_item_with_config(testdir, config, expected_dependencies, expected_targets):
    proj = testdir/'sources/projBatch'
    scheduler_config = SchedulerConfig.from_dict(config)

    # A file with a single subroutine definition that calls two routines via module imports
    item = get_item(
        ProcedureItem, proj/'source/comp2.f90', '#comp2',
        RegexParserClass.ProgramUnitClass, scheduler_config=scheduler_config
    )

    # We need to have suitable dependency modules in the cache to spawn the dependency items
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache.update({
        (i := get_item(
            ModuleItem, proj/path, name,
            RegexParserClass.ProgramUnitClass, scheduler_config=scheduler_config
        )).name: i
        for path, name in [
            ('module/t_mod.F90', 't_mod'), ('module/a_mod.F90', 'a_mod'),
            ('module/b_mod.F90', 'b_mod'), ('headers/header_mod.F90', 'header_mod')
        ]
    })
    assert item.create_dependency_items(item_factory=item_factory, config=scheduler_config) == expected_dependencies


    assert as_tuple(item.disable) == as_tuple(config.get('default', {}).get('disable', []))
    assert item.targets == as_tuple(expected_targets)


@pytest.mark.parametrize('disable', ['#comp2', 'comp2'])
def test_procedure_item_with_config2(testdir, disable):
    proj = testdir/'sources/projBatch'
    scheduler_config = SchedulerConfig.from_dict({'default': {'disable': [disable]}})

    # Similar to the previous test but checking disabling of subroutines without scope
    item = get_item(
        ProcedureItem, proj/'source/comp1.F90', '#comp1',
        RegexParserClass.ProgramUnitClass, scheduler_config=scheduler_config)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache['t_mod'] = get_item(
        ModuleItem, proj/'module/t_mod.F90', 't_mod', RegexParserClass.ProgramUnitClass
    )
    item_factory.item_cache['header_mod'] = get_item(
        ModuleItem, proj/'headers/header_mod.F90', 'header_mod',
        RegexParserClass.ProgramUnitClass, scheduler_config=scheduler_config
    )
    assert item.create_dependency_items(item_factory=item_factory, config=scheduler_config) == (
        't_mod', 't_mod#t', 'header_mod', 't_mod#t%proc', 't_mod#t%no%way'
    )

    assert item.targets == ('t_mod', 't', 'nt1', 'header_mod', 'arg%proc', 'arg%no%way')


@pytest.mark.parametrize('enable_imports', [False, True])
def test_procedure_item_external_item(tmp_path, enable_imports, default_config):
    """
    Test that dependencies to external module procedures are marked as external item
    """
    fcode = """
subroutine procedure_item_external_item
    use external_mod, only: external_proc, unused_external_proc, external_type, external_var
    implicit none
    type(external_type) :: my_type

    call external_proc(1)

    my_type%my_val = external_var
end subroutine procedure_item_external_item
    """
    filepath = tmp_path/'procedure_item_external_item.F90'
    filepath.write_text(fcode)

    default_config['default']['enable_imports'] = enable_imports
    scheduler_config = SchedulerConfig.from_dict(default_config)
    item = get_item(
        ProcedureItem, filepath, '#procedure_item_external_item',
        RegexParserClass.ProgramUnitClass, scheduler_config
    )
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    items = item.create_dependency_items(item_factory=item_factory, config=scheduler_config)

    # NB: dependencies to imported symbols are not added as external items because it would be impossible
    #     to determine their type. Instead, the external module is marked as a dependency, regardless if
    #     imports are enabled or not.
    #     However, the external procedure with a call statement is recognized as an external procedure
    #     and therefore included in the dependency tree.
    assert items == ('external_mod', 'external_mod#external_proc')
    assert all(isinstance(it, ExternalItem) for it in items)
    assert [it.origin_cls for it in items] == [ModuleItem, ProcedureItem]


def test_procedure_item_from_item1(testdir, default_config):
    proj = testdir/'sources/projBatch'

    # A file with a single subroutine definition that calls a routine via interface block
    item_factory = ItemFactory()
    scheduler_config = SchedulerConfig.from_dict(default_config)
    file_item = item_factory.get_or_create_file_item_from_path(proj/'source/comp1.F90', config=scheduler_config)
    item = file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)[0]
    assert item.name == '#comp1'
    assert isinstance(item, ProcedureItem)

    expected_cache = {str(proj/'source/comp1.F90').lower(), '#comp1'}
    assert set(item_factory.item_cache) == expected_cache

    # Create a new item by duplicating the existing item
    new_item = item_factory.get_or_create_item_from_item('#new_comp1', item, config=scheduler_config)
    expected_cache |= {str(proj/'source/new_comp1.F90').lower(), '#new_comp1'}
    assert set(item_factory.item_cache) == expected_cache

    # Assert the new item differs from the existing item in the name, with the original
    # item unchanged
    assert new_item.name == '#new_comp1'
    assert isinstance(new_item, ProcedureItem)
    assert new_item.ir.name == 'new_comp1'
    assert item.ir.name == 'comp1'

    # Make sure both items have the same dependencies but the dependency
    # objects are distinct objects
    assert item.dependencies == new_item.dependencies
    assert all(d is not new_d for d, new_d in zip(item.dependencies, new_item.dependencies))


def test_procedure_item_from_item2(testdir, default_config):
    proj = testdir/'sources/projBatch'

    # A file with a single subroutine declared in a module that calls a typebound procedure
    # where the type is imported via an import statement in the module scope
    item_factory = ItemFactory()
    scheduler_config = SchedulerConfig.from_dict(default_config)
    file_item = item_factory.get_or_create_file_item_from_path(proj/'module/other_mod.F90', config=scheduler_config)
    mod_item = file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)[0]
    assert mod_item.name == 'other_mod'
    assert isinstance(mod_item, ModuleItem)
    item = mod_item.create_definition_items(item_factory=item_factory, config=scheduler_config)[0]
    assert item.name == 'other_mod#mod_proc'
    assert isinstance(item, ProcedureItem)

    expected_cache = {str(proj/'module/other_mod.F90').lower(), 'other_mod', 'other_mod#mod_proc'}
    assert set(item_factory.item_cache) == expected_cache

    # Create a new item by duplicating the existing item
    new_item = item_factory.get_or_create_item_from_item('my_mod#new_proc', item, config=scheduler_config)[0]
    expected_cache |= {str(proj/'module/my_mod.F90').lower(), 'my_mod', 'my_mod#new_proc'}
    assert set(item_factory.item_cache) == expected_cache

    # Assert the new item differs from the existing item in the name, with the original
    # item unchanged
    assert new_item.name == 'my_mod#new_proc'
    assert isinstance(new_item, ProcedureItem)
    assert new_item.ir.name == 'new_proc'
    assert new_item.ir.parent.name == 'my_mod'
    assert item.ir.name == 'mod_proc'
    assert item.ir.parent.name == 'other_mod'

    # Make sure both items have the same dependencies but the dependency
    # objects are distinct objects
    assert item.dependencies == new_item.dependencies
    assert all(d is not new_d for d, new_d in zip(item.dependencies, new_item.dependencies))


def test_typedef_item(testdir):
    proj = testdir/'sources/projBatch'

    # A file with multiple type definitions, of which we pick one
    item = get_item(
        TypeDefItem, proj/'module/t_mod.F90', 't_mod#t',
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass
    )
    assert item.name == 't_mod#t'
    assert str(item) == 'loki.batch.TypeDefItem<t_mod#t>'
    assert item.ir is item.source['t']
    assert 'proc' in item.ir.variable_map
    assert item.definitions == item.ir.declarations

    # Without module items in the cache, the definition items will be externals
    assert all(
        isinstance(_item, ExternalItem) and _item.origin_cls is ProcedureBindingItem
        for _item in item.create_definition_items(item_factory=ItemFactory())
    )
    assert item.dependencies == (item.scope.import_map['tt'], item.ir.parent['t1'])

    # Without module items in the cache, the dependency items will be externals
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    assert all(
        isinstance(_item, ExternalItem) and _item.origin_cls in (ModuleItem, TypeDefItem)
        for _item in item.create_dependency_items(item_factory=ItemFactory())
    )

    # Need to add the modules of the dependent types
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache['t_mod'] = ModuleItem('t_mod', source=item.source)
    item_factory.item_cache['tt_mod'] = get_item(
        ModuleItem, proj/'module/tt_mod.F90', 'tt_mod', RegexParserClass.ProgramUnitClass
    )
    assert 'tt_mod#tt' not in item_factory.item_cache
    assert 't_mod#t1' not in item_factory.item_cache
    items = item.create_dependency_items(item_factory=item_factory)
    assert 'tt_mod#tt' in item_factory.item_cache
    assert 't_mod#t1' in item_factory.item_cache
    assert items == (item_factory.item_cache['tt_mod#tt'], item_factory.item_cache['t_mod#t1'])
    assert all(isinstance(i, TypeDefItem) for i in items[1:])
    assert not items[1].dependencies


def test_interface_item_in_module(testdir):
    proj = testdir/'sources/projInlineCalls'

    # A file containing a module, with an interface to declare multiple functions
    # with a common name
    item = get_item(
        InterfaceItem, proj/'some_module.F90', 'some_module#add_args',
        RegexParserClass.ProgramUnitClass | RegexParserClass.InterfaceClass
    )

    assert item.name == 'some_module#add_args'
    assert str(item) == 'loki.batch.InterfaceItem<some_module#add_args>'
    assert item.ir is item.source['some_module'].interface_map['add_args']
    assert {'add_args', 'add_two_args', 'add_three_args'} == set(item.ir.symbols)

    # An interface does not define anything by itself
    assert not item.definitions
    assert not item.create_definition_items(item_factory=ItemFactory())

    # An interface depends on the routines it declares
    assert item.dependencies == ('add_two_args', 'add_three_args')

    # Without module item in the cache, the dependencies will be externals
    scheduler_config = SchedulerConfig.from_dict({'default': {'strict': True}})
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    assert all(
        isinstance(_item, ExternalItem) and _item.origin_cls is ProcedureItem
        for _item in item.create_dependency_items(item_factory=item_factory, config=scheduler_config)
    )

    # Let's start again with the module item
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache['some_module'] = ModuleItem('some_module', source=item.source)
    assert 'some_module#add_two_args' not in item_factory.item_cache
    assert 'some_module#add_three_args' not in item_factory.item_cache
    items = item.create_dependency_items(item_factory=item_factory)
    assert 'some_module#add_two_args' in item_factory.item_cache
    assert 'some_module#add_three_args' in item_factory.item_cache
    assert items == (
        item_factory.item_cache['some_module#add_two_args'], item_factory.item_cache['some_module#add_three_args']
    )
    assert all(isinstance(i, ProcedureItem) for i in items)


def test_interface_item_in_subroutine(testdir):
    proj = testdir/'sources/projInlineCalls'

    # A file containing the driver subroutine, which uses an interface to declare an
    # inline call
    item = get_item(
        ProcedureItem, proj/'driver.F90', '#driver',
        RegexParserClass.ProgramUnitClass
    )

    # Make sure the interface is included in the dependencies
    assert len(item.dependencies) == len(item.ir.imports + item.ir.interfaces) + 1 # (+1 for the call)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item

    # Dependency items cannot be created without the corresponding modules present
    with pytest.raises(RuntimeError):
        item.create_dependency_items(item_factory=item_factory)

    # Add the missing dependency modules
    for module_name in ('some_module', 'vars_module'):
        module_item = get_item(ModuleItem, proj/f'{module_name}.F90', module_name, RegexParserClass.ProgramUnitClass)
        item_factory.item_cache[module_item.name] = module_item

    # Dependency items can still not be created because the interface routine is still missing
    with pytest.raises(RuntimeError):
        item.create_dependency_items(item_factory=item_factory)

    # Add the missing dependency
    routine_item = get_item(ProcedureItem, proj/'double_real.F90', '#double_real', RegexParserClass.ProgramUnitClass)
    item_factory.item_cache[routine_item.name] = routine_item

    # Validate dependency items
    items = item.create_dependency_items(item_factory=item_factory)
    assert set(items) == {
        'some_module#return_one', 'some_module', 'some_module#add_args', 'some_module#some_type',
        'vars_module', '#double_real', 'some_module#some_type%do_something'
    }


def test_procedure_binding_item1(testdir):
    proj = testdir/'sources/projBatch'
    # pylint: disable=unsupported-binary-operation
    parser_classes = (
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass | RegexParserClass.DeclarationClass
    )

    # A typedef with a procedure binding as well as nested types that have in turn procedure bindings

    # 1. A direct procedure binding
    item = get_item(ProcedureBindingItem, proj/'module/t_mod.F90', 't_mod#t%proc', parser_classes)
    assert item.name == 't_mod#t%proc'
    assert str(item) == 'loki.batch.ProcedureBindingItem<t_mod#t%proc>'
    assert item.ir is item.source['t'].variable_map['proc']
    assert isinstance(item.definitions, tuple) and not item.definitions
    assert not item.create_definition_items(item_factory=ItemFactory())
    assert item.dependencies == as_tuple(item.source['t_proc'])

    item_factory = ItemFactory()
    item_factory.item_cache.update({'t_mod': ModuleItem('t_mod', source=item.source)})
    items = item.create_dependency_items(item_factory=item_factory)
    assert len(items) == 1
    assert isinstance(items[0], ProcedureItem)
    assert items[0].ir is item.source['t_proc']


def test_procedure_binding_item2(testdir, default_config):
    proj = testdir/'sources/projBatch'
    # pylint: disable=unsupported-binary-operation
    parser_classes = (
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass | RegexParserClass.DeclarationClass
    )

    # 2. An indirect procedure binding via a nested type member, where the type is declared in the same module
    item = get_item(ProcedureBindingItem, proj/'module/t_mod.F90', 't_mod#t%no%way', parser_classes)
    assert item.name == 't_mod#t%no%way'
    assert isinstance(item.ir, Scalar)
    assert isinstance(item.definitions, tuple) and not item.definitions
    assert not item.create_definition_items(item_factory=ItemFactory())
    assert item.dependencies == ('no%way',)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    # ExternalItem, because item_cache does not contain the relevant module
    assert all(
        isinstance(_item, ExternalItem) and _item.origin_cls is ProcedureBindingItem
        for _item in item.create_dependency_items(
            item_factory=item_factory, config=SchedulerConfig.from_dict(default_config)
        )
    )

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache['t_mod'] = ModuleItem('t_mod', source=item.source)
    items = item.create_dependency_items(item_factory=item_factory)
    assert len(items) == 1
    assert isinstance(items[0], ProcedureBindingItem)
    assert items[0].name == 't_mod#t1%way'
    assert 't_mod#t1%way' in item_factory.item_cache

    assert 't_mod#my_way' not in item_factory.item_cache
    next_items = items[0].create_dependency_items(item_factory=item_factory)
    assert len(next_items) == 1
    assert isinstance(next_items[0], ProcedureItem)
    assert next_items[0].ir is item.source['my_way']
    assert 't_mod#my_way' in item_factory.item_cache


def test_procedure_binding_item3(testdir):
    proj = testdir/'sources/projBatch'
    # pylint: disable=unsupported-binary-operation
    parser_classes = (
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass | RegexParserClass.DeclarationClass
    )

    # 3. An indirect procedure binding via a nested type member, where the type is declared in a different module
    item = get_item(ProcedureBindingItem, proj/'module/t_mod.F90', 't_mod#t%yay%proc', parser_classes)
    assert item.name == 't_mod#t%yay%proc'
    assert isinstance(item.ir, Scalar)
    assert isinstance(item.definitions, tuple) and not item.definitions
    assert not item.create_definition_items(item_factory=ItemFactory())
    assert item.dependencies == ('yay%proc',)

    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache['tt_mod'] = get_item(ModuleItem, proj/'module/tt_mod.F90', 'tt_mod', parser_classes)
    items = item.create_dependency_items(item_factory=item_factory)
    assert len(items) == 1
    assert isinstance(items[0], ProcedureBindingItem)
    assert items[0].name == 'tt_mod#tt%proc'
    assert 'tt_mod#tt%proc' in item_factory.item_cache

    assert 'tt_mod#proc' not in item_factory.item_cache
    next_items = items[0].create_dependency_items(item_factory=item_factory)
    assert len(next_items) == 1
    assert isinstance(next_items[0], ProcedureItem)
    assert next_items[0].ir is items[0].source['proc']
    assert 'tt_mod#proc' in item_factory.item_cache


@pytest.mark.parametrize('config,expected_dependencies', [
    ({}, (('tt_mod#tt%proc',), ('tt_mod#proc',))),
    ({'default': {'disable': ['tt_mod#proc']}}, (('tt_mod#tt%proc',), ())),
    ({'default': {'disable': ['proc']}}, (('tt_mod#tt%proc',), ())),
    ({'default': {'disable': ['tt%proc']}}, ((),)),
    ({'default': {'disable': ['tt_mod#tt%proc']}}, ((),)),
])
def test_procedure_binding_with_config(testdir, config, expected_dependencies):
    proj = testdir/'sources/projBatch'
    # pylint: disable=unsupported-binary-operation
    parser_classes = (
        RegexParserClass.ProgramUnitClass | RegexParserClass.TypeDefClass | RegexParserClass.DeclarationClass
    )

    item = get_item(ProcedureBindingItem, proj/'module/t_mod.F90', 't_mod#t%yay%proc', parser_classes)

    # We need to have suitable dependency modules in the cache to spawn the dependency items
    item_factory = ItemFactory()
    item_factory.item_cache[item.name] = item
    item_factory.item_cache['tt_mod'] = get_item(
        ModuleItem, proj/'module/tt_mod.F90', 'tt_mod', RegexParserClass.ProgramUnitClass
    )
    scheduler_config = SchedulerConfig.from_dict(config)

    for dependencies in expected_dependencies:
        items = item.create_dependency_items(item_factory, config=scheduler_config)
        assert items == dependencies
        if items:
            item = items[0]


def test_item_graph(testdir, comp1_expected_dependencies):
    """
    Build a :any:`nx.Digraph` from a dummy call hierarchy to check the incremental parsing and
    discovery behaves as expected.
    """
    proj = testdir/'sources/projBatch'
    suffixes = ['.f90', '.F90']

    path_list = [f for ext in suffixes for f in proj.glob(f'**/*{ext}')]
    assert len(path_list) == 8

    # Map item names to items
    item_factory = ItemFactory()

    # Instantiate the basic list of items (files, modules, subroutines)
    for path in path_list:
        relative_path = str(path.relative_to(proj))
        file_item = get_item(FileItem, path, relative_path, RegexParserClass.ProgramUnitClass)
        item_factory.item_cache[relative_path] = file_item
        item_factory.item_cache.update(
            (item.name, item) for item in file_item.create_definition_items(item_factory=item_factory)
        )

    # Populate a graph from a seed routine
    seed = '#comp1'
    queue = deque()
    full_graph = nx.DiGraph()
    full_graph.add_node(item_factory.item_cache[seed])
    queue.append(item_factory.item_cache[seed])

    while queue:
        item = queue.popleft()
        dependencies = item.create_dependency_items(item_factory=item_factory)
        new_items = [i for i in dependencies if i not in full_graph]
        if new_items:
            full_graph.add_nodes_from(new_items)
            queue.extend(new_items)
        full_graph.add_edges_from((item, dependency) for dependency in dependencies)

    # Need to add the cyclic dependency (which isn't included in the fixture)
    comp1_expected_dependencies['t_mod#my_way'] += ('t_mod#t1%way',)

    assert set(full_graph) == set(comp1_expected_dependencies)
    assert {(a.name, b.name) for a, b in full_graph.edges} == {
        (a, b) for a, deps in comp1_expected_dependencies.items() for b in deps
    }

    # Note: quick visualization for debugging can be done using matplotlib
    # import matplotlib.pyplot as plt
    # nx.draw_planar(full_graph, with_labels=True)
    # plt.show()
    # # -or-
    # plt.savefig('test_item_graph.png')


@pytest.mark.parametrize('seed,dependencies_fixture', [
    ('#comp1', 'comp1_expected_dependencies'),
    ('other_mod#mod_proc', 'mod_proc_expected_dependencies'),
    (['#comp1', 'other_mod#mod_proc'], 'expected_dependencies'),
    ('#foobar', 'no_expected_dependencies'),
    # Not fully-qualified procedure name for a free subroutine
    ('comp1', 'comp1_expected_dependencies'),
     # Not fully-qualified procedure name for a module procedure
    ('mod_proc', 'mod_proc_expected_dependencies'),
])
def test_sgraph_from_seed(tmp_path, testdir, default_config, seed, dependencies_fixture, request):
    expected_dependencies = request.getfixturevalue(dependencies_fixture)
    proj = testdir/'sources/projBatch'
    suffixes = ['.f90', '.F90']

    path_list = [f for ext in suffixes for f in proj.glob(f'**/*{ext}')]
    assert len(path_list) == 8

    scheduler_config = SchedulerConfig.from_dict(default_config)
    item_factory = ItemFactory()

    # Instantiate the basic list of items (files, modules, subroutines)
    for path in path_list:
        relative_path = str(path.relative_to(proj))
        file_item = get_item(
            FileItem, path, relative_path, RegexParserClass.ProgramUnitClass,
            scheduler_config
        )
        item_factory.item_cache[relative_path] = file_item
        item_factory.item_cache.update(
            (item.name, item)
            for item in file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)
        )

    # Create the graph
    sgraph = SGraph.from_seed(seed, item_factory, scheduler_config)

    # Check the graph
    assert set(sgraph.items) == set(expected_dependencies)
    assert set(sgraph.dependencies) == {
        (node, dependency)
        for node, dependencies in expected_dependencies.items()
        for dependency in dependencies
    }

    # Check the graph visualization
    graph_file = tmp_path/'sgraph_from_seed.dot'
    sgraph.export_to_file(graph_file)
    assert graph_file.exists()
    assert graph_file.with_suffix('.dot.pdf').exists()

    vgraph = VisGraphWrapper(graph_file)
    assert set(vgraph.nodes) == {item.upper() for item in expected_dependencies}
    assert set(vgraph.edges) == {
        (node.upper(), dependency.upper())
        for node, dependencies in expected_dependencies.items()
        for dependency in dependencies
    }


@pytest.mark.parametrize('seed,disable,active_nodes', [
    ('#comp1', ('comp2', 'a'), (
        '#comp1', 't_mod', 't_mod#t', 'header_mod', 't_mod#t%proc', 't_mod#t%no%way',
        't_mod#t_proc', 't_mod#t%yay%proc', 'tt_mod#tt%proc', 'tt_mod#proc',
        't_mod#t1%way', 't_mod#my_way', 'tt_mod#tt', 't_mod#t1'
    )),
    ('#comp1', ('comp2', 'a', 't_mod#t%no%way'), (
        '#comp1', 't_mod', 't_mod#t', 'header_mod', 't_mod#t%proc',
        't_mod#t_proc', 't_mod#t%yay%proc', 'tt_mod#tt%proc', 'tt_mod#proc',
        'tt_mod#tt', 't_mod#t1'
    )),
    ('#comp1', ('#comp2', 't1%way'), (
        '#comp1', 't_mod', 't_mod#t', 'header_mod', 't_mod#t%proc', 't_mod#t%no%way',
        't_mod#t_proc', 't_mod#t%yay%proc', 'tt_mod#tt%proc', 'tt_mod#proc',
        'tt_mod#tt', 't_mod#t1', 'a_mod#a'
    )),
    ('t_mod#t_proc', ('t_mod#t1', 'proc'), (
        't_mod#t_proc', 't_mod#t', 'tt_mod#tt', 'a_mod#a', 'header_mod',
        't_mod#t%yay%proc', 'tt_mod#tt%proc'
    ))
])
def test_sgraph_disable(testdir, default_config, expected_dependencies, seed, disable, active_nodes):
    proj = testdir/'sources/projBatch'
    suffixes = ['.f90', '.F90']

    path_list = [f for ext in suffixes for f in proj.glob(f'**/*{ext}')]
    assert len(path_list) == 8

    default_config['default']['disable'] = disable
    scheduler_config = SchedulerConfig.from_dict(default_config)
    item_factory = ItemFactory()

    # Instantiate the basic list of items (files, modules, subroutines)
    for path in path_list:
        relative_path = str(path.relative_to(proj))
        file_item = get_item(
            FileItem, path, relative_path, RegexParserClass.ProgramUnitClass,
            scheduler_config
        )
        item_factory.item_cache[relative_path] = file_item
        item_factory.item_cache.update(
            (item.name, item)
            for item in file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)
        )

    # Create the graph
    sgraph = SGraph.from_seed(seed, item_factory, scheduler_config)

    # Check the graph
    assert set(sgraph.items) == set(active_nodes)
    assert set(sgraph.dependencies) == {
        (node, dependency)
        for node, dependencies in expected_dependencies.items()
        for dependency in dependencies
        if node in active_nodes and dependency in active_nodes
    }


@pytest.mark.parametrize('seed,routines,active_nodes', [
    (
        '#comp1', {
            'comp1': {'expand': False}
        }, (
            '#comp1',
        )
    ),
    (
        '#comp2', {
            'comp2': {'block': ['a', 'b']},
            't_mod': {'block': ['a']}
        }, (
            '#comp2', 't_mod#t', 'header_mod', 't_mod#t%yay%proc',
            'tt_mod#tt', 't_mod#t1', 'tt_mod#tt%proc', 'tt_mod#proc'
        )
    ),
    (
        '#comp2', {
            'comp2': {'ignore': ['a'], 'block': ['b']},
            't_mod': {'ignore': ['a']}
        }, (
            '#comp2', 't_mod#t', 'header_mod', 't_mod#t%yay%proc',
            'tt_mod#tt', 't_mod#t1', 'tt_mod#tt%proc', 'tt_mod#proc',
            'a_mod#a'
        )
    ),
])
def test_sgraph_routines(testdir, default_config, expected_dependencies, seed, routines, active_nodes):
    proj = testdir/'sources/projBatch'
    suffixes = ['.f90', '.F90']

    path_list = [f for ext in suffixes for f in proj.glob(f'**/*{ext}')]
    assert len(path_list) == 8

    default_config['routines'] = routines
    scheduler_config = SchedulerConfig.from_dict(default_config)
    item_factory = ItemFactory()

    # Instantiate the basic list of items (files, modules, subroutines)
    for path in path_list:
        relative_path = str(path.relative_to(proj))
        file_item = get_item(
            FileItem, path, relative_path, RegexParserClass.ProgramUnitClass,
            scheduler_config
        )
        item_factory.item_cache[relative_path] = file_item
        item_factory.item_cache.update(
            (item.name, item)
            for item in file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)
        )

    # Create the graph
    sgraph = SGraph.from_seed(seed, item_factory, scheduler_config)

    # Check the graph
    assert set(sgraph.items) == set(active_nodes)
    assert set(sgraph.dependencies) == {
        (node, dependency)
        for node, dependencies in expected_dependencies.items()
        for dependency in dependencies
        if node in active_nodes and dependency in active_nodes
    }

    targets = expected_dependencies[seed]
    targets = [t.replace('t_mod#t%', 'arg%') for t in targets]
    targets = [t.rsplit('#', maxsplit=1)[-1] for t in targets]

    # Without full parse and enriching (as done in the Scheduler before processing),
    # the type of the imported symbol cannot be determined and therefore global
    # variables like `nt1` or parameters like 'k' are listed as targets
    if seed == '#comp1':
        targets += ['nt1']
    if seed == '#comp2':
        targets += ['t_mod', 'b_mod', 'a_mod', 'k']

    if 'block' in routines[seed[1:]]:
        targets = [t for t in targets if t not in routines[seed[1:]]['block']]
    assert set(item_factory.item_cache[seed].targets) == set(targets)

    item_factory.item_cache['t_mod'].source.make_complete()
    item_factory.item_cache['header_mod'].source.make_complete()
    item_factory.item_cache[seed].source.make_complete()
    item_factory.item_cache[seed].ir.enrich([
        item_factory.item_cache['t_mod'].ir, item_factory.item_cache['header_mod'].ir]
    )

    # With fully-parsed and enriched source, we are able to distinguish between
    # the types of imported symbols and consequently global variables and parameters
    # are no longer listed as targets
    if 'nt1' in targets:
        targets.remove('nt1')
    if 'k' in targets:
        targets.remove('k')
    assert set(item_factory.item_cache[seed].targets) == set(targets)


def test_sgraph_filegraph(testdir, default_config, file_dependencies):
    proj = testdir/'sources/projBatch'
    suffixes = ['.f90', '.F90']

    path_list = [f for ext in suffixes for f in proj.glob(f'**/*{ext}')]
    assert len(path_list) == 8

    scheduler_config = SchedulerConfig.from_dict(default_config)
    item_factory = ItemFactory()

    # Instantiate the basic list of items (files, modules, subroutines)
    for path in path_list:
        file_item = FileItem(
            name=str(path),
            source=Sourcefile.from_file(path, frontend=REGEX, parser_classes=RegexParserClass.ProgramUnitClass),
            config=scheduler_config.create_item_config(str(path))
        )
        item_factory.item_cache[file_item.name] = file_item
        item_factory.item_cache.update(
            (item.name, item)
            for item in file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)
        )

    # Create the graph
    sgraph = SGraph.from_seed('#comp1', item_factory, scheduler_config)

    # Derive the file graph
    file_graph = SGraph.as_filegraph(sgraph, item_factory, scheduler_config)

    assert set(file_graph.items) == {str(proj/name) for name in file_dependencies}
    assert set(file_graph.dependencies) == {
        (str(proj/node), str(proj/dependency))
        for node, dependencies in file_dependencies.items()
        for dependency in dependencies
    }

def discover_proj_typebound_item_factory(testdir, scheduler_config):
    proj = testdir/'sources/projTypeBound'
    suffixes = ['.f90', '.F90']

    path_list = [f for ext in suffixes for f in proj.glob(f'**/*{ext}')]
    assert len(path_list) == 3

    item_factory = ItemFactory()

    # Instantiate the basic list of items (files, modules, subroutines)
    for path in path_list:
        file_item = FileItem(
            name=str(path),
            source=Sourcefile.from_file(path, frontend=REGEX, parser_classes=RegexParserClass.ProgramUnitClass),
        )
        item_factory.item_cache[file_item.name] = file_item

        definitions = {
            item.name: item
            for item in file_item.create_definition_items(item_factory=item_factory, config=scheduler_config)
        }
        item_factory.item_cache.update(definitions)

        module_names = [item.name for item in definitions.values() if isinstance(item, ModuleItem)]
        definitions = {
            item.name: item
            for module in module_names
            for item in item_factory.item_cache[module].create_definition_items(
                item_factory=item_factory, config=scheduler_config
            )
        }
        item_factory.item_cache.update(definitions)

        type_names = [item.name for item in definitions.values() if isinstance(item, TypeDefItem)]
        definitions = {
            item.name: item
            for type_ in type_names
            for item in item_factory.item_cache[type_].create_definition_items(
                item_factory=item_factory, config=scheduler_config
            )
        }
        item_factory.item_cache.update(definitions)

    return item_factory


@pytest.mark.parametrize('name,config_override,item_type,ir_type,attrs_to_check,dependency_items', [
    (
        ##########
        '#driver',
        ##########
        # This depends on  modules (via unqualified imports), typebound procedures,
        # and typebound procedures in nested derived types
        {},
        ProcedureItem,
        Subroutine,
        {
            'calls': (
                'some_type%other_routine',
                'some_type%some_routine',
                'header_type%member_routine',
                'header_type%routine',
                'other%member',
                'other%var%member_routine'
            ),
            'targets': (
                'typebound_item',
                'typebound_header',
                'typebound_other',
                'other',
                'obj%other_routine',
                'obj2%some_routine',
                'header%member_routine',
                'header%routine',
                'other_obj%member',
                'derived%var%member_routine'
            )
        },
        (
            'typebound_item',
            'typebound_header',
            'typebound_other#other_type',
            'typebound_item#some_type%other_routine',
            'typebound_item#some_type%some_routine',
            'typebound_header#header_type%member_routine',
            'typebound_header#header_type%routine',
            'typebound_other#other_type%member',
            'typebound_other#other_type%var%member_routine'
        )
    ),
    (
        ###############################
        'typebound_item#other_routine',
        ###############################
        # This is a module routine that depends on a subroutine from an unqualified import,
        # and typebound procedures in the same module
        {},
        ProcedureItem,
        Subroutine,
        {
            'calls': ('abor1', 'some_type%routine1', 'some_type%routine2'),
            'targets': ('some_type', 'abor1', 'self%routine1', 'self%routine2'),
        },
        (
            'typebound_item#some_type',
            'typebound_header#abor1',
            'typebound_item#some_type%routine1',
            'typebound_item#some_type%routine2'
        ),
    ),
    (
        #########################
        'typebound_item#routine',
        #########################
        # This is a module routine that depends on a type bound procedure,
        # which is listed as disabled in the scheduler config with fully qualified name
        {
            'disable': ['typebound_item#some_type%some_routine'],
        },
        ProcedureItem,
        Subroutine,
        {
            'calls': ('some_type%some_routine',),
            'targets': ('some_type',)
        },
        ('typebound_item#some_type',),
    ),
    (
        #########################
        'typebound_item#routine',
        #########################
        # This is a module routine that depends on a type bound procedure,
        # which is listed as disabled in the scheduler config without providing scope
        {
            'disable': ['some_type%some_routine'],
        },
        ProcedureItem,
        Subroutine,
        {
            'calls': ('some_type%some_routine',),
            'targets': ('some_type',)
        },
        ('typebound_item#some_type',),
    ),
    (
        #########################
        'typebound_item#routine1',
        #########################
        # This is a module routine that depends on a module procedure,
        # which is listed as disabled in the scheduler config without providing scope
        {
            'disable': ['module_routine'],
        },
        ProcedureItem,
        Subroutine,
        {
            'calls': ('module_routine',),
            'targets': ('some_type',)
        },
        ('typebound_item#some_type',),
    ),
    (
        #######################################
        'typebound_item#some_type%some_routine',
        #######################################
        # This is a procedure binding, where the bound procedure is listed as disabled
        # without providing fully qualified name - This means that the dependency item
        # is not created
        {
            'disable': ['some_routine'],
        },
        ProcedureBindingItem,
        ProcedureSymbol,
        {
            'calls': (),
            'targets': (),
        },
        (),
    ),
    (
        #######################################
        'typebound_item#some_type%some_routine',
        #######################################
        # This is a procedure binding, where the bound procedure is listed as disabled
        # with fully qualified name provided - this means that the dependency item
        # is not created
        {
            'disable': ['typebound_item#some_routine'],
        },
        ProcedureBindingItem,
        ProcedureSymbol,
        {
            'calls': (),
            'targets': (),
        },
        (),
    ),
    (
        #######################################
        'typebound_item#some_type%some_routine',
        #######################################
        # This is a procedure binding, where the bound procedure is listed as ignored,
        # which still includes it in the targets list
        {
            'ignore': ['some_routine'],
        },
        ProcedureBindingItem,
        ProcedureSymbol,
        {
            'calls': (),
            'targets': ('some_routine',)
        },
        ('typebound_item#some_routine',),
    ),
    (
        #######################################
        'typebound_item#some_type%some_routine',
        #######################################
        # This is a procedure binding, where the bound procedure is listed as blocked,
        # which excludes it from the targets list
        {
            'block': ['some_routine'],
        },
        ProcedureBindingItem,
        ProcedureSymbol,
        {
            'calls': (),
            'targets': ()
        },
        (),
    ),
    (
        ###################################
        'typebound_item#some_type%routine',
        ###################################
        # This is a procedure binding with renaming
        {},
        ProcedureBindingItem,
        ProcedureSymbol,
        {
            'calls': (),
            'targets': ('module_routine',),
        },
        ('typebound_item#module_routine',),
    ),
    (
        #############################
        'typebound_other#other_type',
        #############################
        # This is a derived type definition that has a dependency on another
        # type that is imported from another module, and renamed upon import
        {},
        TypeDefItem,
        TypeDef,
        {
            'calls': (),
            'targets': ('typebound_header', 'header')
        },
        ('typebound_header#header_type',),
    ),
])
def test_batch_typebound_item(
    testdir, default_config,
    name, config_override, item_type, ir_type, attrs_to_check, dependency_items
):
    """
    Test the basic regex frontend nodes in :any:`Item` objects for fast dependency detection
    for type-bound procedures.
    """
    default_config['default'].update(config_override)
    scheduler_config = SchedulerConfig.from_dict(default_config)
    item_factory = discover_proj_typebound_item_factory(testdir, scheduler_config)

    item = item_factory.item_cache[name]
    assert isinstance(item, item_type)
    assert isinstance(item.ir, ir_type)

    for key, value in attrs_to_check.items():
        assert getattr(item, key) == value

    assert item.create_dependency_items(item_factory, scheduler_config) == dependency_items


def test_batch_typebound_nested_item(testdir, default_config):
    """
    Test the basic regex frontend nodes in :any:`Item` objects for fast dependency detection
    for type-bound procedures for calls to nested derived type bindings
    """
    scheduler_config = SchedulerConfig.from_dict(default_config)
    item_factory = discover_proj_typebound_item_factory(testdir, scheduler_config)

    item = item_factory.item_cache['typebound_other#other_member']
    assert isinstance(item, ProcedureItem)
    assert isinstance(item.ir, Subroutine)
    assert len(item.dependencies) == 4

    assert isinstance(item.dependencies[0], ir.Import)
    assert item.dependencies[0].module == 'typebound_header'

    # Verify that the call to the nested type's routine is added when creating
    # dependency items
    assert 'typebound_other#other_type%var%member_routine' not in item_factory
    assert item.create_dependency_items(item_factory, scheduler_config) == (
        'typebound_other#other_type',
        'typebound_header#header_member_routine',
        'typebound_other#other_type%var%member_routine',
    )
    assert 'typebound_other#other_type%var%member_routine' in item_factory

    # Verify that the nested binding item can correctly resolve this to the binding
    # in the type
    proc_bind_item = item_factory.item_cache['typebound_other#other_type%var%member_routine']
    assert isinstance(proc_bind_item, ProcedureBindingItem)
    assert isinstance(proc_bind_item.ir, Scalar)
    assert proc_bind_item.ir == 'var'
    assert proc_bind_item.dependencies == ('var%member_routine',)
    assert proc_bind_item.create_dependency_items(item_factory, scheduler_config) == (
        'typebound_header#header_type%member_routine',
    )

    # Verify that the binding in the type correctly resolves to the module routine
    nested_bind_item = item_factory.item_cache['typebound_header#header_type%member_routine']
    assert isinstance(nested_bind_item, ProcedureBindingItem)
    assert isinstance(nested_bind_item.ir, ProcedureSymbol)
    assert nested_bind_item.ir == 'member_routine'
    assert nested_bind_item.create_dependency_items(item_factory, scheduler_config) == (
        'typebound_header#header_member_routine',
    )

    # Verify that we're now at the module routine
    routine_item = item_factory.item_cache['typebound_header#header_member_routine']
    assert isinstance(routine_item, ProcedureItem)
    assert isinstance(routine_item.ir, Subroutine)

    # Lastly, look at the deeply nested call...
    nested_call_item = item_factory.item_cache['typebound_other#nested_call']
    assert isinstance(nested_call_item, ProcedureItem)
    assert nested_call_item.create_dependency_items(item_factory, scheduler_config) == (
        'typebound_other#outer_type', 'typebound_other#outer_type%other%var%member_routine'
    )

    # ...and see if we can chase the deeply nested dependencies correctly
    other_var_member_item = item_factory.item_cache['typebound_other#outer_type%other%var%member_routine']
    assert isinstance(other_var_member_item, ProcedureBindingItem)
    assert isinstance(other_var_member_item.ir, Scalar)
    assert other_var_member_item.dependencies == ('other%var%member_routine',)
    assert other_var_member_item.create_dependency_items(item_factory, scheduler_config) == (
        'typebound_other#other_type%var%member_routine',
    )


def test_batch_typebound_item_targets(default_config):
    default_config['default']['disable'] += ['timer_mod']

    fcode = """
MODULE TYPEBOUND_ITEM_TARGETS_MOD
    USE TIMER_MOD, ONLY: PERFORMANCE_TIMER
    IMPLICIT NONE
CONTAINS
    SUBROUTINE DRIVER
        IMPLICIT NONE
        TYPE(PERFORMANCE_TIMER) :: TIMER

        CALL TIMER%START()

        ! DO SOMETHING

        CALL TIMER%END()
    END SUBROUTINE DRIVER
END MODULE TYPEBOUND_ITEM_TARGETS_MOD
    """.strip()

    source = Sourcefile.from_source(fcode, parser_classes=RegexParserClass.ProgramUnitClass, frontend=REGEX)
    source.path = 'None'
    item_factory = ItemFactory()
    scheduler_config = SchedulerConfig.from_dict(default_config)

    file_item = item_factory.get_or_create_file_item_from_source(source, scheduler_config)
    assert file_item.targets == ()
    assert file_item.definitions == (source['typebound_item_targets_mod'],)
    assert file_item.dependencies == ()

    file_definitions = file_item.create_definition_items(item_factory, scheduler_config)
    assert file_definitions == ('typebound_item_targets_mod',)

    module_item = file_definitions[0]
    assert module_item.targets == ()
    assert module_item.definitions == (source['driver'],)
    assert module_item.dependencies == module_item.ir.imports

    module_definitions = module_item.create_definition_items(item_factory, scheduler_config)
    assert module_definitions == ('typebound_item_targets_mod#driver',)
    assert module_item.create_dependency_items(item_factory, scheduler_config) == ()

    driver_item = module_definitions[0]
    assert driver_item.targets == ()
    assert driver_item.definitions == ()
    assert ('timer_mod', 'timer%start', 'timer%end') == (
        (driver_item.dependencies[0].module.lower(),) +
        tuple(dep.name for dep in driver_item.dependencies[1:])
    )

    assert driver_item.create_definition_items(item_factory, scheduler_config) == ()
    assert driver_item.create_dependency_items(item_factory, scheduler_config) == ()



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
import importlib
from pathlib import Path
import pytest

from conftest import run_linter, available_frontends
from loki import Sourcefile, FindInlineCalls, FindNodes, VariableDeclaration
from loki.lint import DefaultHandler


pytestmark = pytest.mark.skipif(not available_frontends(),
                                reason='Supported frontend not available')


@pytest.fixture(scope='module', name='rules')
def fixture_rules():
    rules = importlib.import_module('lint_rules.debug_rules')
    return rules


@pytest.mark.parametrize('frontend', available_frontends())
def test_arg_size_array_slices(rules, frontend):
    """
    Test for argument size mismatch when arguments are passed as array slices.
    """

    fcode_driver = """
subroutine driver(klon, klev, nblk, var0, var1, var2, var3, var4, var5, &
                  var6, var7)
use yomhook, only : lhook,   dr_hook, jphook
implicit none

integer, intent(in) :: klon, klev, nblk
real, intent(in) :: var2(:,:), var4(:,:), var5(:,:), var3(klon, 137), var5(klon, 138)
real, intent(in) :: var6(:,:), var7(:,:)
real, intent(inout) :: var0(klon, nblk), var1(klon, 138, nblk)
real(kind=jphook) :: zhook_handle
integer :: klev, ibl, iproma, iend

if(lhook) call dr_hook('driver', 0, zhook_handle)

associate(nlev => klev)
nlev = 137
do ibl = 1, nblk
   iproma = klon
   iend = iproma
   call kernel(klon, nlev, var0(:,ibl), var1(:,:,ibl), var2(1:iend, 1:nlev), &
               var3, var4(1:klon, 1:nlev+1), var5(:, 1:nlev+1), &
               var6_d=var6, var7_d=var7(:,1:nlev))
enddo
end associate

if(lhook) call dr_hook('driver', 1, zhook_handle)
end subroutine driver
    """.strip()

    fcode_kernel = """
subroutine kernel(klon, klev, var0_d, var1_d, var2_d, var3_d, var4_d, var5_d, var6_d, var7_d)
use yomhook, only : lhook,   dr_hook, jphook
implicit none
integer, intent(in) :: klon, klev
real, dimension(klon, klev), intent(inout) :: var0_d, var1_d
real, dimension(klon, klev), intent(in) :: var2_d, var3_d, var4_d
real, dimension(klon, klev+1), intent(in) :: var5_d
real, intent(in) :: var6_d(klon, klev), var7_d(klon, klev)
real(kind=jphook) :: zhook_handle

if(lhook) call dr_hook('kernel', 0, zhook_handle)
if(lhook) call dr_hook('kernel', 1, zhook_handle)
end subroutine kernel
    """.strip()

    driver_source = Sourcefile.from_source(fcode_driver, frontend=frontend)
    kernel_source = Sourcefile.from_source(fcode_kernel, frontend=frontend)

    driver = driver_source['driver']
    kernel = kernel_source['kernel']
    driver.enrich([kernel,])

    messages = []
    handler = DefaultHandler(target=messages.append)
    _ = run_linter(driver_source, [rules.ArgSizeMismatchRule], config={'ArgSizeMismatchRule': {'max_indirections': 3}},
                   handlers=[handler], targets=['kernel',])

    assert len(messages) == 3
    keyword = 'ArgSizeMismatchRule'
    assert all(keyword in msg for msg in messages)

    args = ('var0', 'var1', 'var4')
    for msg, ref_arg in zip(messages, args):
        assert f'arg: {ref_arg}' in msg
        assert f'dummy_arg: {ref_arg}_d' in msg


@pytest.mark.parametrize('frontend', available_frontends())
def test_arg_size_array_sequence(rules, frontend):
    """
    Test for argument size mismatch when arguments are passed as array sequences.
    """

    fcode_driver = """
subroutine driver(klon, klev, nblk, var0, var1, var2, var3)
use yomhook, only : lhook,   dr_hook, jphook
implicit none

integer, intent(in) :: klon, klev, nblk
real, intent(inout) :: var0(klon, nblk), var1(klon, 138, nblk)
real, intent(in) ::  var2(klon, 137), var3(klon*137)
real(kind=jphook) :: zhook_handle
real, dimension(klon, 137) :: var4, var5
real :: var6
integer :: klev, ibl, iproma, iend

if(lhook) call dr_hook('driver', 0, zhook_handle)

klev = 137
do ibl = 1, nblk
   iproma = klon
   iend = iproma
   call kernel(klon, klev, var0(1,ibl), var1(1,1,ibl), var2(1:iend, 1), var3(1), &
               var4(1, 1), var5, var6, 1, .true.)
enddo

if(lhook) call dr_hook('driver', 1, zhook_handle)
end subroutine driver
    """.strip()

    fcode_kernel = """
subroutine kernel(klon, klev, var0_d, var1_d, var2_d, var3_d, var4_d, var5_d, var6_d, &
                  int_arg, log_arg)
use yomhook, only : lhook,   dr_hook, jphook
implicit none
integer, intent(in) :: klon, klev
real, dimension(klon, klev), intent(inout) :: var0_d, var1_d
real, dimension(klon, klev), intent(in) :: var2_d, var3_d
real, intent(out) :: var4_d, var5_d, var6_d(klon, klev)
integer, intent(out) :: int_arg
logical, intent(out) :: log_arg
real(kind=jphook) :: zhook_handle

if(lhook) call dr_hook('kernel', 0, zhook_handle)
if(lhook) call dr_hook('kernel', 1, zhook_handle)
end subroutine kernel
    """.strip()

    driver_source = Sourcefile.from_source(fcode_driver, frontend=frontend)
    kernel_source = Sourcefile.from_source(fcode_kernel, frontend=frontend)

    driver = driver_source['driver']
    kernel = kernel_source['kernel']
    driver.enrich([kernel,])

    messages = []
    handler = DefaultHandler(target=messages.append)
    _ = run_linter(driver_source, [rules.ArgSizeMismatchRule], handlers=[handler], targets=['kernel',])

    assert len(messages) == 4
    keyword = 'ArgSizeMismatchRule'
    assert all(keyword in msg for msg in messages)

    args = ('var0', 'var1', 'var5', 'var6')
    for msg, ref_arg in zip(messages, args):
        assert f'arg: {ref_arg}' in msg
        assert f'dummy_arg: {ref_arg}_d' in msg


@pytest.mark.parametrize('frontend', available_frontends())
def test_dynamic_ubound_checks(rules, frontend):
    """
    Test the run-time UBOUND checking linter rule
    """

    fcode = """
subroutine kernel(klon, klev, nblk, var0, var1, var2, var3, var4)
use abort_mod
implicit none
integer, intent(in) :: klon, klev, nblk
real, dimension(:,:,:), intent(inout) :: var0, var1
real, dimension(:,:,:), intent(inout) :: var2
real, intent(inout) :: var3(:,:), var4(:,:,:)

if(ubound(var0, 1) < klon)then
  call abort('kernel: first dimension of var0 too short')
endif
if(ubound(VAR0, 2) < klev)then
  call abort('kernel: second dimension of var0 too short')
endif
if(nblk > UBoUND(vAr0, 3))then
  call abort('kernel: third dimension of var0 too short')
endif

if(nblk > UBOUND(var1, 3))then
  call abort('kernel: third dimension of var1 too short')
endif

if(ubound(var2, 1) < klon .and. ubound(var2, 2) < klev .and. ubound(var2, 3) < nblk)then
  call abort('kernel: dimensions of var2 too short')
endif

if(ubound(var4, 1) < klon .and. ubound(var4, 2) < klev .and. ubound(var4, 3) < nblk)then
  call abort('kernel: dimensions of var4 too short')
endif

call some_other_kernel(klon, klen, nblk, var0, var1, var2, var3, var4)

end subroutine kernel
    """.strip()

    kernel = Sourcefile.from_source(fcode, frontend=frontend)
    kernel.path = Path(__file__).parent / 'dynamic_ubound_test.F90'

    messages = []
    handler = DefaultHandler(target=messages.append)
    _ = run_linter(kernel, [rules.DynamicUboundCheckRule], config={'fix': True}, handlers=[handler])

    # check rule violations
    assert len(messages) == 3
    assert all('DynamicUboundCheckRule' in msg for msg in messages)

    assert 'var0' in messages[0]
    assert 'var2' in messages[1]
    assert 'var4' in messages[2]

    # check fixed subroutine
    routine = kernel['kernel']
    icalls = [call for call in FindInlineCalls(unique=False).visit(routine.body)
              if call.function == 'ubound']

    assert len(icalls) == 1

    shape = ('klon', 'klev', 'nblk')

    assert all(s.name == d for s, d in zip(routine.variable_map['var0'].shape, shape))
    assert all(s.name == d for s, d in zip(routine.variable_map['var2'].shape, shape))
    assert all(s.name == d for s, d in zip(routine.variable_map['var4'].shape, shape))

    arg_names = ['klon', 'klev', 'nblk', 'var0', 'var1', 'var2', 'var3', 'var4']
    assert [arg.name.lower() for arg in routine.arguments] == arg_names

    # check that variable declarations have not been duplicated
    declarations = FindNodes(VariableDeclaration).visit(routine.spec)
    symbols = [s.name.lower() for decl in declarations for s in decl.symbols]
    assert len(symbols) == 8
    assert set(symbols) == {'klon', 'klev', 'nblk', 'var0', 'var1', 'var2', 'var3', 'var4'}

    # check number of declarations and symbols per declarations
    assert len(declarations) == 5
    assert len(declarations[0].symbols) == 3
    for decl in declarations[1:4]:
        assert len(decl.symbols) == 1
    assert len(declarations[4].symbols) == 2

    os.remove(kernel.path)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Collection of utility routines that provide transformations for code regions.

"""
from collections import defaultdict

from loki.ir import (
    Comment, Loop, Pragma, PragmaRegion, FindNodes, FindScopes,
    MaskedTransformer, NestedMaskedTransformer, is_loki_pragma,
    get_pragma_parameters, pragma_regions_attached
)
from loki.logging import info
from loki.tools import as_tuple, flatten

from loki.transformations.array_indexing import (
    promotion_dimensions_from_loop_nest, promote_nonmatching_variables
)


__all__ = ['region_hoist']


def region_hoist(routine):
    """
    Hoist one or multiple code regions annotated by pragma ranges and insert
    them at a specified target location.

    The pragma syntax for annotating the regions to hoist is
    ``!$loki region-hoist [group(group-name)] [collapse(n) [promote(var-name, var-name, ...)]]``
    and ``!$loki end region-hoist``.
    The optional ``group(group-name)`` can be provided when multiple regions
    are to be hoisted and inserted at different positions. Multiple pragma
    ranges can be specified for the same group, all of which are then moved to
    the target location in the same order as the pragma ranges appear.
    The optional ``collapse(n)`` parameter specifies that ``n`` enclosing scopes
    (such as loops, conditionals, etc.) should be re-created at the target location.
    Optionally, this can be combined with variable promotion using ``promote(...)``.
    """
    hoist_targets = defaultdict(list)
    hoist_regions = defaultdict(list)

    # Find all region-hoist pragma regions
    with pragma_regions_attached(routine):
        for region in FindNodes(PragmaRegion).visit(routine.body):
            if is_loki_pragma(region.pragma, starts_with='region-hoist'):
                parameters = get_pragma_parameters(region.pragma, starts_with='region-hoist')
                group = parameters.get('group', 'default')
                hoist_regions[group] += [(region.pragma, region.pragma_post)]

    # Find all region-hoist targets
    for pragma in FindNodes(Pragma).visit(routine.body):
        if is_loki_pragma(pragma, starts_with='region-hoist'):
            parameters = get_pragma_parameters(pragma, starts_with='region-hoist')
            if 'target' in parameters:
                group = parameters.get('group', 'default')
                hoist_targets[group] += [pragma]

    if not hoist_regions:
        return

    # Group-by-group extract the regions and build the node replacement map
    hoist_map = {}
    promotion_vars_dims = {}  # Variables to promote with new dimension
    promotion_vars_index = {}  # Variable subscripts to promote with new indices
    starts, stops = [], []
    for group, regions in hoist_regions.items():
        if not group in hoist_targets or not hoist_targets[group]:
            raise RuntimeError(f'No region-hoist target for group {group} defined.')
        if len(hoist_targets[group]) > 1:
            raise RuntimeError(f'Multiple region-hoist targets given for group {group}')

        hoist_body = ()
        for start, stop in regions:
            parameters = get_pragma_parameters(start, starts_with='region-hoist')

            # Extract the region to hoist
            collapse = int(parameters.get('collapse', 0))
            if collapse > 0:
                scopes = FindScopes(start).visit(routine.body)[0]
                if len(scopes) <= collapse:
                    raise RuntimeError(f'Not enough enclosing scopes for collapse({collapse})')
                scopes = scopes[-(collapse+1):]
                region = NestedMaskedTransformer(start=start, stop=stop, mapper={start: None}).visit(scopes[0])

                # Promote variables given in promotion list
                loops = [scope for scope in scopes if isinstance(scope, Loop)]
                promote_vars = [var.strip().lower()
                                for var in get_pragma_parameters(start).get('promote', '').split(',') if var]
                promotion_vars_dims, promotion_vars_index = promotion_dimensions_from_loop_nest(
                    promote_vars, loops, promotion_vars_dims, promotion_vars_index)
            else:
                region = MaskedTransformer(start=start, stop=stop, mapper={start: None}).visit(routine.body)

            # Append it to the group's body, wrapped in comments
            begin_comment = Comment(f'! Loki {start.content}')
            end_comment = Comment(f'! Loki {stop.content}')
            hoist_body += as_tuple(flatten([begin_comment, region, end_comment]))

            # Register start and end nodes for transformer mask
            starts += [stop]
            stops += [start]

            # Replace end pragma by comment
            comment = Comment(f'! Loki {start.content} - region hoisted')
            hoist_map[stop] = comment

        # Insert target <-> hoisted regions into map
        hoist_map[hoist_targets[group][0]] = hoist_body

    routine.body = MaskedTransformer(active=True, start=starts, stop=stops, mapper=hoist_map).visit(routine.body)
    num_targets = sum(1 for pragma in hoist_map if 'target' in get_pragma_parameters(pragma))
    info('%s: hoisted %d region(s) in %d group(s)', routine.name, len(hoist_map) - num_targets, num_targets)
    promote_nonmatching_variables(routine, promotion_vars_dims, promotion_vars_index)



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
Contains the declaration of :any:`Sourcefile` that is used to represent and
manipulate (Fortran) source code files.
"""
from pathlib import Path
from codetiming import Timer

from loki.backend.fgen import fgen
from loki.backend.cufgen import cufgen
from loki.frontend import (
    Frontend, OMNI, FP, REGEX, sanitize_input, Source, read_file,
    preprocess_cpp, parse_omni_source, parse_fparser_source,
    parse_omni_ast, parse_fparser_ast, parse_regex_source,
    RegexParserClass

)
from loki.ir import Section, RawSource, Comment, PreprocessorDirective
from loki.logging import debug, detail, perf
from loki.module import Module
from loki.program_unit import ProgramUnit
from loki.subroutine import Subroutine
from loki.tools import flatten, as_tuple


__all__ = ['Sourcefile']


class Sourcefile:
    """
    Class to handle and manipulate source files, storing :any:`Module` and
    :any:`Subroutine` objects.

    Reading existing source code from file or string can be done via
    :meth:`from_file` or :meth:`from_source`.

    Parameters
    ----------
    path : str
        The name of the source file.
    ir : :any:`Section`, optional
        The IR of the file content (including :any:`Subroutine`, :any:`Module`,
        :any:`Comment` etc.)
    ast : optional
        Parser-AST of the original source file.
    source : :any:`Source`, optional
        Raw source string and line information about the original source file.
    incomplete : bool, optional
        Mark the object as incomplete, i.e. only partially parsed. This is
        typically the case when it was instantiated using the :any:`Frontend.REGEX`
        frontend and a full parse using one of the other frontends is pending.
    parser_classes : :any:`RegexParserClass`, optional
        Provide the list of parser classes used during incomplete regex parsing
    """

    def __init__(self, path, ir=None, ast=None, source=None, incomplete=False, parser_classes=None):
        self.path = Path(path) if path is not None else path
        if ir is not None and not isinstance(ir, Section):
            ir = Section(body=ir)
        self.ir = ir
        self._ast = ast
        self._source = source
        self._incomplete = incomplete
        self._parser_classes = parser_classes

    def clone(self, **kwargs):
        """
        Replicate the object with the provided overrides.
        """
        kwargs.setdefault('path', self.path)
        if self.ir is not None and 'ir' not in kwargs:
            kwargs['ir'] = self.ir
            ir_needs_clone = True
        else:
            ir_needs_clone = False
        if self._ast is not None and 'ast' not in kwargs:
            kwargs['ast'] = self._ast
        if self.source is not None and 'source' not in kwargs:
            kwargs['source'] = self._source.clone(file=kwargs['path'])
        kwargs.setdefault('incomplete', self._incomplete)
        if self._parser_classes is not None and 'parser_classes' not in kwargs:
            kwargs['parser_classes'] = self._parser_classes

        obj = type(self)(**kwargs)

        # When the IR has been carried over from the current sourcefile
        # we need to make sure we perform a deep copy
        if obj.ir and ir_needs_clone:
            ir_body = tuple(
                node.clone(rescope_symbols=True) if isinstance(node, ProgramUnit)
                else node.clone() for node in obj.ir.body
            )
            obj.ir = obj.ir.clone(body=ir_body)
        return obj

    @classmethod
    def from_file(cls, filename, definitions=None, preprocess=False,
                  includes=None, defines=None, omni_includes=None,
                  xmods=None, frontend=FP, parser_classes=None):
        """
        Constructor from raw source files that can apply a
        C-preprocessor before invoking frontend parsers.

        Parameters
        ----------
        filename : str
            Name of the file to parse into a :any:`Sourcefile` object.
        definitions : list of :any:`Module`, optional
            :any:`Module` object(s) that may supply external type or procedure
            definitions.
        preprocess : bool, optional
            Flag to trigger CPP preprocessing (by default `False`).

            .. attention::
                Please note that, when using the OMNI frontend, C-preprocessing
                will always be applied, so :data:`includes` and :data:`defines`
                may have to be defined even when disabling :data:`preprocess`.

        includes : list of str, optional
            Include paths to pass to the C-preprocessor.
        defines : list of str, optional
            Symbol definitions to pass to the C-preprocessor.
        xmods : str, optional
            Path to directory to find and store ``.xmod`` files when using the
            OMNI frontend.
        omni_includes: list of str, optional
            Additional include paths to pass to the preprocessor run as part of
            the OMNI frontend parse. If set, this **replaces** (!)
            :data:`includes`, otherwise :data:`omni_includes` defaults to the
            value of :data:`includes`.
        frontend : :any:`Frontend`, optional
            Frontend to use for producing the AST (default :any:`FP`).
        """
        if isinstance(frontend, str):
            frontend = Frontend[frontend.upper()]

        # Log full parses at INFO and regex scans at DETAIL level
        log = f'[Loki::Sourcefile] Constructed from {filename}' + ' in {:.2f}s'
        with Timer(logger=detail if frontend is REGEX else perf, text=log):

            filepath = Path(filename)
            raw_source = read_file(filepath)

            if preprocess:
                # Trigger CPP-preprocessing explicitly, as includes and
                # defines can also be used by our OMNI frontend
                source = preprocess_cpp(source=raw_source, filepath=filepath,
                                        includes=includes, defines=defines)
            else:
                source = raw_source

            if frontend == REGEX:
                return cls.from_regex(source, filepath, parser_classes=parser_classes)

            if frontend == OMNI:
                return cls.from_omni(source, filepath, definitions=definitions,
                                     includes=includes, defines=defines,
                                     xmods=xmods, omni_includes=omni_includes)

            if frontend == FP:
                return cls.from_fparser(source, filepath, definitions=definitions)

            raise NotImplementedError(f'Unknown frontend: {frontend}')

    @classmethod
    def from_omni(cls, raw_source, filepath, definitions=None, includes=None,
                  defines=None, xmods=None, omni_includes=None):
        """
        Parse a given source string using the OMNI frontend

        Parameters
        ----------
        raw_source : str
            Fortran source string
        filepath : str or :any:`pathlib.Path`
            The filepath of this source file
        definitions : list
            List of external :any:`Module` to provide derived-type and procedure declarations
        includes : list of str, optional
            Include paths to pass to the C-preprocessor.
        defines : list of str, optional
            Symbol definitions to pass to the C-preprocessor.
        xmods : str, optional
            Path to directory to find and store ``.xmod`` files when using the
            OMNI frontend.
        omni_includes: list of str, optional
            Additional include paths to pass to the preprocessor run as part of
            the OMNI frontend parse. If set, this **replaces** (!)
            :data:`includes`, otherwise :data:`omni_includes` defaults to the
            value of :data:`includes`.
        """
        # Always CPP-preprocess source files for OMNI, but optionally
        # use a different set of include paths if specified that way.
        # (It's a hack, I know, but OMNI sucks, so what can I do...?)
        if omni_includes is not None and len(omni_includes) > 0:
            includes = omni_includes
        source = preprocess_cpp(raw_source, filepath=filepath,
                                includes=includes, defines=defines)

        # Parse the file content into an OMNI Fortran AST
        ast = parse_omni_source(source=source, filepath=filepath, xmods=xmods)
        typetable = ast.find('typeTable')
        return cls._from_omni_ast(ast=ast, path=filepath, raw_source=raw_source,
                                  definitions=definitions, typetable=typetable)

    @classmethod
    def _from_omni_ast(cls, ast, path=None, raw_source=None, definitions=None, typetable=None):
        """
        Generate the full set of `Subroutine` and `Module` members of the `Sourcefile`.
        """
        type_map = {t.attrib['type']: t for t in typetable}
        if ast.find('symbols') is not None:
            symbol_map = {s.attrib['type']: s for s in ast.find('symbols')}
        else:
            symbol_map = None

        ir = parse_omni_ast(
            ast=ast, definitions=definitions, raw_source=raw_source,
            type_map=type_map, symbol_map=symbol_map
        )

        lines = (1, raw_source.count('\n') + 1)
        source = Source(lines, string=raw_source, file=path)
        return cls(path=path, ir=ir, ast=ast, source=source)

    @classmethod
    def from_fparser(cls, raw_source, filepath, definitions=None):
        """
        Parse a given source string using the fparser frontend

        Parameters
        ----------
        raw_source : str
            Fortran source string
        filepath : str or :any:`pathlib.Path`
            The filepath of this source file
        definitions : list
            List of external :any:`Module` to provide derived-type and procedure declarations
        """
        # Preprocess using internal frontend-specific PP rules
        # to sanitize input and work around known frontend problems.
        source, pp_info = sanitize_input(source=raw_source, frontend=FP)

        # Parse the file content into a Fortran AST
        ast = parse_fparser_source(source)

        return cls._from_fparser_ast(path=filepath, ast=ast, definitions=definitions,
                                     pp_info=pp_info, raw_source=raw_source)

    @classmethod
    def _from_fparser_ast(cls, ast, path=None, raw_source=None, definitions=None, pp_info=None):
        """
        Generate the full set of :any:`Subroutine` and :any:`Module` members
        in the :any:`Sourcefile`.
        """
        ir = parse_fparser_ast(ast, pp_info=pp_info, definitions=definitions, raw_source=raw_source)

        lines = (1, raw_source.count('\n') + 1)
        source = Source(lines, string=raw_source, file=path)
        return cls(path=path, ir=ir, ast=ast, source=source)

    @classmethod
    def from_regex(cls, raw_source, filepath, parser_classes=None):
        """
        Parse a given source string using the REGEX frontend
        """
        source, _ = sanitize_input(source=raw_source, frontend=REGEX)

        if parser_classes is None:
            parser_classes = RegexParserClass.AllClasses
        ir = parse_regex_source(source, parser_classes=parser_classes)
        lines = (1, raw_source.count('\n') + 1)
        source = Source(lines, string=raw_source, file=filepath)
        return cls(path=filepath, ir=ir, source=source, incomplete=True, parser_classes=parser_classes)

    @classmethod
    def from_source(cls, source, definitions=None, preprocess=False,
                    includes=None, defines=None, omni_includes=None,
                    xmods=None, frontend=FP, parser_classes=None):
        """
        Constructor from raw source string that invokes specified frontend parser

        Parameters
        ----------
        source : str
            Fortran source string
        definitions : list of :any:`Module`, optional
            :any:`Module` object(s) that may supply external type or procedure
            definitions.
        preprocess : bool, optional
            Flag to trigger CPP preprocessing (by default `False`).

            .. attention::
                Please note that, when using the OMNI frontend, C-preprocessing
                will always be applied, so :data:`includes` and :data:`defines`
                may have to be defined even when disabling :data:`preprocess`.

        includes : list of str, optional
            Include paths to pass to the C-preprocessor.
        defines : list of str, optional
            Symbol definitions to pass to the C-preprocessor.
        xmods : str, optional
            Path to directory to find and store ``.xmod`` files when using the
            OMNI frontend.
        omni_includes: list of str, optional
            Additional include paths to pass to the preprocessor run as part of
            the OMNI frontend parse. If set, this **replaces** (!)
            :data:`includes`, otherwise :data:`omni_includes` defaults to the
            value of :data:`includes`.
        frontend : :any:`Frontend`, optional
            Frontend to use for producing the AST (default :any:`FP`).
        """
        if isinstance(frontend, str):
            frontend = Frontend[frontend.upper()]

        if preprocess:
            # Trigger CPP-preprocessing explicitly, as includes and
            # defines can also be used by our OMNI frontend
            source = preprocess_cpp(source=source, includes=includes, defines=defines)

        if frontend == REGEX:
            return cls.from_regex(source, filepath=None, parser_classes=parser_classes)

        if frontend == OMNI:
            return cls.from_omni(source, filepath=None, definitions=definitions, includes=includes,
                                 defines=defines, xmods=xmods, omni_includes=omni_includes)

        if frontend == FP:
            return cls.from_fparser(source, filepath=None, definitions=definitions)

        raise NotImplementedError(f'Unknown frontend: {frontend}')

    def make_complete(self, **frontend_args):
        """
        Trigger a re-parse of the source file if incomplete to produce a full Loki IR

        If the source file is marked to be incomplete, i.e. when using the `lazy` constructor
        option, this triggers a new parsing of all :any:`ProgramUnit` objects and any
        :any:`RawSource` nodes in the :attr:`Sourcefile.ir`.

        Existing :any:`Module` and :any:`Subroutine` objects continue to exist and references
        to them stay valid, as they will only be updated instead of replaced.
        """
        if not self._incomplete:
            return

        frontend = frontend_args.pop('frontend', FP)

        log = f'[Loki::Sourcefile] Finished constructing from {self.path}' + ' in {:.2f}s'
        with Timer(logger=debug if frontend == REGEX else perf, text=log):

            # Sanitize frontend_args
            if isinstance(frontend, str):
                frontend = Frontend[frontend.upper()]
            if frontend == REGEX:
                frontend_argnames = ['parser_classes']
            elif frontend == OMNI:
                frontend_argnames = ['definitions', 'type_map', 'symbol_map', 'scope']
                xmods = frontend_args.get('xmods')
            elif frontend == FP:
                frontend_argnames = ['definitions', 'scope']
            else:
                raise NotImplementedError(f'Unknown frontend: {frontend}')
            sanitized_frontend_args = {k: frontend_args.get(k) for k in frontend_argnames}

            body = []
            for node in self.ir.body:
                if isinstance(node, ProgramUnit):
                    node.make_complete(frontend=frontend, **frontend_args)
                    body += [node]
                elif isinstance(node, RawSource):
                    # Sanitize the input code to ensure non-supported features
                    # do not break frontend parsing ourside of program units
                    raw_source = node.source.string
                    source, pp_info = sanitize_input(source=raw_source, frontend=frontend)

                    # Typically, this should only be comments, PP statements etc., therefore
                    # we are not bothering with type tables, definitions or similar to parse them
                    if frontend == REGEX:
                        ir_ = parse_regex_source(source, **sanitized_frontend_args)
                    elif frontend == OMNI:
                        ast = parse_omni_source(source=source, xmods=xmods)
                        ir_ = parse_omni_ast(ast=ast, raw_source=raw_source, **sanitized_frontend_args)
                    elif frontend == FP:
                        # Fparser is unable to parse comment-only source files/strings,
                        # so we see if this is only comments and convert them ourselves
                        # (https://github.com/stfc/fparser/issues/375)
                        # This can be removed once fparser 0.0.17 is released
                        lines = [l.lstrip() for l in source.splitlines()]
                        if all(not l or l[0] in '!#' for l in lines):
                            ir_ = [
                                PreprocessorDirective(text=line.string, source=line)
                                if line.string.lstrip().startswith('#')
                                else Comment(text=line.string, source=line)
                                for line in node.source.clone_lines()
                            ]
                        else:
                            ast = parse_fparser_source(source)
                            ir_ = parse_fparser_ast(ast, raw_source=raw_source, pp_info=pp_info,
                                                    **sanitized_frontend_args)
                    else:
                        raise NotImplementedError(f'Unknown frontend: {frontend}')
                    if isinstance(ir_, Section):
                        ir_ = ir_.body
                    body += flatten([ir_])
                else:
                    body += [node]

            self.ir._update(body=as_tuple(body))
            self._incomplete = frontend == REGEX
            if frontend == REGEX:
                parser_classes = frontend_args.get('parser_classes', RegexParserClass.AllClasses)
                if self._parser_classes:
                    parser_classes = self._parser_classes | parser_classes
                self._parser_classes = parser_classes

    @property
    def source(self):
        return self._source

    def to_fortran(self, conservative=False, cuf=False):
        if cuf:
            return cufgen(self, conservative=conservative)
        return fgen(self, conservative=conservative)

    @property
    def modules(self):
        """
        List of :class:`Module` objects that are members of this :class:`Sourcefile`.
        """
        if self.ir is None:
            return ()
        return as_tuple(
            module for module in self.ir.body if isinstance(module, Module)
        )

    @property
    def routines(self):
        """
        List of :class:`Subroutine` objects that are members of this :class:`Sourcefile`.
        """
        if self.ir is None:
            return ()
        return as_tuple(
            routine for routine in self.ir.body if isinstance(routine, Subroutine)
        )

    subroutines = routines

    @property
    def typedefs(self):
        """
        List of :class:`TypeDef` objects that are declared in the :any:`Module` in this :class:`Sourcefile`.
        """
        if self.ir is None:
            return ()
        return as_tuple(flatten(module.typedefs for module in self.modules))

    @property
    def all_subroutines(self):
        routines = self.subroutines
        routines += as_tuple(flatten(m.subroutines for m in self.modules))
        return routines

    @property
    def definitions(self):
        """
        List of all definitions made in this sourcefile, i.e. modules and subroutines
        """
        return self.modules + self.subroutines

    def __contains__(self, name):
        """
        Check if a module, type, or subroutine with the given name is declared
        inside this sourcefile
        """
        return self[name] is not None

    def __getitem__(self, name):
        name = name.lower()
        for module in self.modules:
            if name == module.name.lower():
                return module

        for routine in self.all_subroutines:
            if name == routine.name.lower():
                return routine

        for module in self.modules:
            for typedef in module.typedefs:
                if name == typedef.name.lower():
                    return typedef
            for interface in module.interfaces:
                if name in interface.symbols:
                    return interface

        return None

    def __iter__(self):
        raise TypeError('Sourcefiles alone cannot be traversed! Try traversing "Sourcefile.ir".')

    def __bool__(self):
        """
        Ensure existing objects register as True in boolean checks, despite
        raising exceptions in `__iter__`.
        """
        return True

    @property
    def _canonical(self):
        """
        Base definition for comparing :any:`Subroutine` objects.
        """
        return (self.path, self.ir, self.source, )

    def __eq__(self, other):
        if isinstance(other, Sourcefile):
            return self._canonical == other._canonical
        return super().__eq__(other)

    def __hash__(self):
        return hash(self._canonical)

    def __getstate__(self):
        # Do not pickle the AST, as it is not pickle-safe for certain frontends
        _ignore = ('_ast',)
        return dict((k, v) for k, v in self.__dict__.items() if k not in _ignore)

    def apply(self, op, **kwargs):
        """
        Apply a given transformation to the source file object.

        Note that the dispatch routine `op.apply(source)` will ensure
        that all entities of this `Sourcefile` are correctly traversed.
        """
        # TODO: Should type-check for an `Operation` object here
        op.apply(self, **kwargs)

    def write(self, path=None, source=None, conservative=False, cuf=False):
        """
        Write content as Fortran source code to file

        Parameters
        ----------
        path : str, optional
            Filepath of target file; if not provided, :attr:`Sourcefile.path` is used
        source : str, optional
            Write the provided string instead of generating via :any:`Sourcefile.to_fortran`
        conservative : bool, optional
            Enable conservative output in the backend, aiming to be as much string-identical
            as possible (default: False)
        cuf: bool, optional
            To use either Cuda Fortran or Fortran backend
        """
        path = self.path if path is None else Path(path)
        source = self.to_fortran(conservative, cuf) if source is None else source
        self.to_file(source=source, path=path)

    @classmethod
    def to_file(cls, source, path):
        """
        Same as :meth:`write` but can be called from a static context.
        """
        detail(f'[Loki::Sourcefile] Writing to {path}')
        with path.open('w') as f:
            f.write(source)
            if source[-1] != '\n':
                f.write('\n')



# (C) Copyright 2018- ECMWF.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from loki import HAVE_FP, FP
from loki.lint import Reporter, Linter


__all__ = ['available_frontends', 'run_linter']


def available_frontends():
    """Choose frontend to use (Linter currently relies exclusively on Fparser)"""
    if HAVE_FP:
        return [FP,]
    return []


def run_linter(sourcefile, rule_list, config=None, handlers=None, targets=None):
    """
    Run the linter for the given source file with the specified list of rules.
    """
    reporter = Reporter(handlers)
    linter = Linter(reporter, rules=rule_list, config=config)
    report = linter.check(sourcefile, targets=targets)
    if config:
        if config.get('fix', None):
            linter.fix(sourcefile, report)
    return linter
